# coding=utf-8
#
# Copyright 2015-2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""BIG-IP® Application Security Manager™ (ASM®) tasks sub-module.

REST URI
    ``http://localhost/mgmt/tm/asm/tasks/``

GUI Path
    ``Security``

REST Kind
    ``tm:asm:tasks:``
"""
from f5.bigip.resource import AsmResource
from f5.bigip.resource import AsmTaskResource
from f5.bigip.resource import Collection
from f5.bigip.resource import OrganizingCollection
from f5.sdk_exception import F5SDKError
from f5.sdk_exception import RequiredOneOf
from f5.sdk_exception import UnsupportedOperation
from six import iterkeys


class Tasks(OrganizingCollection):
    """BIG-IP® ASM Tasks organizing collection."""
    def __init__(self, asm):
        super(Tasks, self).__init__(asm)
        self._meta_data['object_has_stats'] = False
        self._meta_data['allowed_lazy_attributes'] = [
            Apply_Policy_s,
            Check_Signatures_s,
            Export_Policy_s,
            Export_Signatures_s,
            Import_Policy_s,
            Import_Vulnerabilities_s,
            Update_Signatures_s,
        ]


class Apply_Policy_s(Collection):
    """BIG-IP® ASM Apply Policy Collection."""
    def __init__(self, tasks):
        super(Apply_Policy_s, self).__init__(tasks)
        self._meta_data['object_has_stats'] = False
        self._meta_data['allowed_lazy_attributes'] = [Apply_Policy]
        self._meta_data['attribute_registry'] = {
            'tm:asm:tasks:apply-policy:apply-policy-taskstate': Apply_Policy}


class Apply_Policy(AsmResource):
    """BIG-IP® ASM Apply Policy Resource."""
    def __init__(self, apply_policy_s):
        super(Apply_Policy, self).__init__(apply_policy_s)
        self._meta_data['required_json_kind'] =\
            'tm:asm:tasks:apply-policy:apply-policy-taskstate'
        self._meta_data['required_creation_parameters'] = {'policyReference', }

    def modify(self, **kwargs):
        """Modify is not supported for Apply Policy resource

                :raises: UnsupportedOperation
        """
        raise UnsupportedOperation(
            "%s does not support the modify method" % self.__class__.__name__
        )


class Export_Policy_s(Collection):
    """BIG-IP® ASM Export Policy Collection."""
    def __init__(self, tasks):
        super(Export_Policy_s, self).__init__(tasks)
        self._meta_data['object_has_stats'] = False
        self._meta_data['allowed_lazy_attributes'] = [Export_Policy]
        self._meta_data['attribute_registry'] = {
            'tm:asm:tasks:export-policy:export-policy-taskstate':
                Export_Policy}


class Export_Policy(AsmResource):
    """BIG-IP® ASM Export Policy Resource."""
    def __init__(self, export_policy_s):
        super(Export_Policy, self).__init__(export_policy_s)
        self._meta_data['required_json_kind'] =\
            'tm:asm:tasks:export-policy:export-policy-taskstate'
        self._meta_data['required_creation_parameters'] = {
            'policyReference'}
        self._meta_data['exclusive_attributes'] = ['filename', 'inline']
        self._meta_data['minimum_additional_parameters'] = {'filename', 'inline'}

    def modify(self, **kwargs):
        """Modify is not supported for Apply Policy resource

                :raises: UnsupportedOperation
        """
        raise UnsupportedOperation(
            "%s does not support the modify method" % self.__class__.__name__
        )


class Import_Policy_s(Collection):
    """BIG-IP® ASM Import Policy Collection."""
    def __init__(self, tasks):
        super(Import_Policy_s, self).__init__(tasks)
        self._meta_data['object_has_stats'] = False
        self._meta_data['allowed_lazy_attributes'] = [Import_Policy]
        self._meta_data['attribute_registry'] = {
            'tm:asm:tasks:import-policy:import-policy-taskstate':
                Import_Policy}


class Import_Policy(AsmResource):
    """BIG-IP® ASM Import Policy Resource."""
    def __init__(self, import_policy_s):
        super(Import_Policy, self).__init__(import_policy_s)
        self._meta_data['required_json_kind'] = 'tm:asm:tasks:import-policy:import-policy-taskstate'
        self._meta_data['required_creation_parameters'] = set()
        self._meta_data['exclusive_attributes'] = ['filename', 'file']
        self._meta_data['minimum_additional_parameters'] = {'filename', 'file', 'policyTemplateReference'}

    def create(self, **kwargs):
        required_one_of = ['name', 'fullPath']
        has_any = [x for x in iterkeys(kwargs) if x in required_one_of]

        if 'partition' in kwargs and 'name' in kwargs and 'fullPath' not in kwargs:
            partition = kwargs.pop('partition', None)
            name = kwargs.pop('name', None)
            if partition is not None and not partition.startswith('/'):
                kwargs['fullPath'] = '/{0}/{1}'.format(partition, name)
            elif partition is None:
                raise F5SDKError(
                    "The partition name, if specified, cannot be none"
                )
            elif partition.startswith('/'):
                kwargs['fullPath'] = '{0}/{1}'.format(partition, name)

        if len(has_any) == 1:
            return self._create(**kwargs)

        raise RequiredOneOf(required_one_of)

    def modify(self, **kwargs):
        """Modify is not supported for Apply Policy resource

                :raises: UnsupportedOperation
        """
        raise UnsupportedOperation(
            "%s does not support the modify method" % self.__class__.__name__
        )


class Check_Signatures_s(Collection):
    """BIG-IP® ASM Tasks Check Signatures Collection."""
    def __init__(self, tasks):
        super(Check_Signatures_s, self).__init__(tasks)
        self._meta_data['object_has_stats'] = False
        self._meta_data['allowed_lazy_attributes'] = [Check_Signature]
        self._meta_data['attribute_registry'] = {
            'tm:asm:tasks:check-signatures:check-signatures-taskstate':
                Check_Signature}


class Check_Signature(AsmTaskResource):
    """BIG-IP® ASM Tasks Check Signature Resource


        To create this resource on the ASM, one must utilize fetch() method
        from AsmTaskResource class, create() is not supported.
    """
    def __init__(self, check_signatures_s):
        super(Check_Signature, self).__init__(check_signatures_s)
        self._meta_data['required_json_kind'] =\
            'tm:asm:tasks:check-signatures:check-signatures-taskstate'


class Export_Signatures_s(Collection):
    """BIG-IP® ASM Tasks Export Signatures Collection."""
    def __init__(self, tasks):
        super(Export_Signatures_s, self).__init__(tasks)
        self._meta_data['object_has_stats'] = False
        self._meta_data['allowed_lazy_attributes'] = [Export_Signature]
        self._meta_data['attribute_registry'] = {
            'tm:asm:tasks:export-signatures:export-signatures-taskstate':
                Export_Signature}


class Export_Signature(AsmResource):
    """BIG-IP® ASM Tasks Export Signature Resource"""
    def __init__(self, export_signatures_s):
        super(Export_Signature, self).__init__(export_signatures_s)
        self._meta_data['required_json_kind'] =\
            'tm:asm:tasks:export-signatures:export-signatures-taskstate'
        self._meta_data['required_creation_parameters'] = {'filename', }

    def modify(self, **kwargs):
        """Modify is not supported for Export Signature resource

                :raises: UnsupportedOperation
        """
        raise UnsupportedOperation(
            "%s does not support the modify method" % self.__class__.__name__
        )


class Update_Signatures_s(Collection):
    """BIG-IP® ASM Tasks Update Signatures Collection."""
    def __init__(self, tasks):
        super(Update_Signatures_s, self).__init__(tasks)
        self._meta_data['object_has_stats'] = False
        self._meta_data['minimum_version'] = '12.0.0'
        self._meta_data['allowed_lazy_attributes'] = [Update_Signature]
        self._meta_data['attribute_registry'] = {
            'tm:asm:tasks:update-signatures:update-signatures-taskstate': Update_Signature
        }


class Update_Signature(AsmTaskResource):
    """BIG-IP® ASM Tasks Update Signature Resource resource

        To create this resource on the ASM, one must utilize fetch() method
        from AsmTaskResource class, create() is not supported.
    """
    def __init__(self, update_signatures_s):
        super(Update_Signature, self).__init__(update_signatures_s)
        self._meta_data['minimum_version'] = '12.0.0'
        self._meta_data['required_json_kind'] = 'tm:asm:tasks:update-signatures:update-signatures-taskstate'


class Import_Vulnerabilities_s(Collection):
    """BIG-IP® ASM Import Vulnerabilities Collection."""
    def __init__(self, tasks):
        super(Import_Vulnerabilities_s, self).__init__(tasks)
        self._meta_data['object_has_stats'] = False
        self._meta_data['minimum_version'] = '11.6.0'
        self._meta_data['allowed_lazy_attributes'] = [Import_Vulnerabilities]
        self._meta_data['attribute_registry'] = {
            'tm:asm:tasks:import-vulnerabilities:'
            'import-vulnerabilities-taskstate': Import_Vulnerabilities}


class Import_Vulnerabilities(AsmResource):
    """BIG-IP® ASM Import Vulnerabilities Resource."""
    def __init__(self, import_vulnerabilities_s):
        super(Import_Vulnerabilities, self).__init__(import_vulnerabilities_s)
        self._meta_data['required_json_kind'] = \
            'tm:asm:tasks:import-vulnerabilities:' \
            'import-vulnerabilities-taskstate'
        self._meta_data['required_creation_parameters'] = {
            'policyReference', 'filename'}
        self._meta_data['minimum_additional_parameters'] = {
            'onlyGetDomainNames', 'importAllDomainNames', 'domainNames'}

    def modify(self, **kwargs):
        """Modify is not supported for Import Vulnerabilities resource

                :raises: UnsupportedOperation
        """
        raise UnsupportedOperation(
            "%s does not support the modify method" % self.__class__.__name__
        )



# coding=utf-8
#
# Copyright 2015 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""BIG-IQ® System (sys) module

REST URI
    ``http://localhost/mgmt/tm/sys/``

GUI Path
    ``System``

REST Kind
    ``tm:sys:*``
"""

from f5.bigiq.resource import OrganizingCollection
from f5.bigiq.tm.sys.dns import Dns
from f5.bigiq.tm.sys.global_settings import Global_Settings
from f5.bigiq.tm.sys.httpd import Httpd
from f5.bigiq.tm.sys.software import Software
from f5.bigiq.tm.sys.sshd import Sshd


class Sys(OrganizingCollection):
    def __init__(self, tm):
        super(Sys, self).__init__(tm)
        self._meta_data['allowed_lazy_attributes'] = [
            Dns,
            Global_Settings,
            Httpd,
            Software,
            Sshd
        ]



# coding=utf-8
#
#  Copyright 2017 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""BIG-IP® Advanced Firewall Manager™ (AFM®) module.

REST URI
    ``http://localhost/mgmt/tm/security/protocol-inspection``

GUI Path
    ``Security --> Protocol Security --> Inspection_Profiles
      Security --> Protocol Security --> Inspection_List
    ``

REST Kind
    ``tm:security:protocol-inspection*``
"""

from f5.bigip.resource import Collection
from f5.bigip.resource import OrganizingCollection
from f5.bigip.resource import Resource
from f5.sdk_exception import UnsupportedMethod


class Protocol_Inspection(OrganizingCollection):
    """BIG-IP® Protocol Inspection Organizing collection"""
    def __init__(self, security):
        super(Protocol_Inspection, self).__init__(security)
        self._meta_data['allowed_lazy_attributes'] = [
            Profiles,
            Compliances,
            Signatures,
        ]


class Profiles(Collection):
    """"BIG-IP® Protocol Inspection Profile collection"""
    def __init__(self, protocol_inspection):
        super(Profiles, self).__init__(protocol_inspection)
        self._meta_data['allowed_lazy_attributes'] = [Profile]
        self._meta_data['attribute_registry'] = \
            {'tm:security:protocol-inspection:profile:profilestate':
                Profile}


class Profile(Resource):
    """BIG-IP® Protocol Inspection Profile resource"""
    def __init__(self, profiles):
        super(Profile, self).__init__(profiles)
        self._meta_data['required_json_kind'] = \
            'tm:security:protocol-inspection:profile:profilestate'
        self._meta_data['required_creation_parameters'].update(('partition',))


class Compliances(Collection):
    def __init__(self, protocol_inspection):
        super(Compliances, self).__init__(protocol_inspection)
        self._meta_data['allowed_lazy_attributes'] = [Compliance]
        self._meta_data['attribute_registry'] = \
            {'tm:security:protocol-inspection:compliance:compliancestate':
                Compliance}


class Compliance(Resource):
    """BIG-IP® Protocol Inspection Compliance resource"""
    def __init__(self, compliances):
        super(Compliance, self).__init__(compliances)
        self._meta_data['required_json_kind'] = \
            'tm:security:protocol-inspection:compliance:compliancestate'
        self._meta_data['required_load_parameters'] = set()

    def create(self, **kwargs):
        raise UnsupportedMethod(
            "%s does not support the create method" % self.__class__.__name__
        )

    def delete(self, **kwargs):
        raise UnsupportedMethod(
            "%s does not support the delete method" % self.__class__.__name__
        )

    def modify(self, **kwargs):
        raise UnsupportedMethod(
            "%s does not support the delete method" % self.__class__.__name__
        )


class Signatures(Collection):
    """BIG-IP® Protocol Inspection Signature collection"""
    def __init__(self, protocol_inspection):
        super(Signatures, self).__init__(protocol_inspection)
        self._meta_data['allowed_lazy_attributes'] = [Signature]
        self._meta_data['attribute_registry'] = \
            {'tm:security:protocol-inspection:signature:signaturestate':
                Signature}


class Signature(Resource):
    """BIG-IP® Protocol Inspection Signature resource"""
    def __init__(self, signatures):
        super(Signature, self).__init__(signatures)
        self._meta_data['required_json_kind'] = \
            'tm:security:protocol-inspection:signature:signaturestate'
        self._meta_data['required_creation_parameters'].update((
            'partition', 'sig', 'description'))



# coding=utf-8
#
# Copyright 2014 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""BIG-IP® system dns module

REST URI
    ``http://localhost/mgmt/tm/transaction``

REST Kind
    ``tm:transaction*``
"""

from f5.bigip.resource import Collection
from f5.bigip.resource import Resource


class Transactions(Collection):
    """This class is a context manager for iControl transactions.

    Upon successful exit of the with statement, the transaction will be
    submitted, otherwise it will be rolled back.

    NOTE: This feature was added to BIGIP in version 11.0.0.

    Example:
    > bigip = BigIP(<args>)
    > tx = bigip.transactions.transaction
    > with TransactionContextManager(tx) as api:
    >     api.net.pools.pool.create(name="foo")
    >     api.sys.dbs.db.update(name="setup.run", value="false")
    >     <perform actions inside a transaction>
    >
    > # transaction is committed when you exit the "with" statement.
    """
    def __init__(self, api):
        super(Transactions, self).__init__(api)
        self._meta_data['allowed_lazy_attributes'] = [Transaction]
        self._meta_data['attribute_registry'] = \
            {'tm:transactionstate': Transaction}


class Transaction(Resource):
    def __init__(self, transactions):
        super(Transaction, self).__init__(transactions)
        self._meta_data['required_json_kind'] = 'tm:transactionstate'
        self._meta_data['required_creation_parameters'] = set()



# coding=utf-8
#
# Copyright 2014 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import copy
import logging

from f5.sdk_exception import TransactionSubmitException


class TransactionContextManager(object):
    def __init__(self, transaction, validate_only=False):
        """Initialize a new Transaction context

        Args:
            validate_only (bool): Will not commit the transaction, but only
                validate that it will succeed or not.

        Attributes:
            transaction (Transaction): The transaction object that was sent
                to the context manager
            validate_only (bool): Specifies whether the transaction should
                commit itself upon `__exit__` or whether the commands in the
                transaction should just be checked to make sure they don't
                raise an error.
            bigip (dict): A reference to the dictionary containing the BIG-IP
                mgmt_root
            icr (iControlRESTSession): A reference to the dictionary
                containing the iControl REST session
            original_headers (dict): A deep copy of all the headers that were
                originally part of the iControl REST session. A copy is needed
                so that we can revert back to them after the transaction has
                been committed, since the only way to commit the transaction
                is to set the X-F5-REST-Coordination-Id to the value of the
                transaction ID of the transaction.
        """
        self.transaction = transaction
        self.validate_only = validate_only
        self.bigip = transaction._meta_data['bigip']
        self.icr = self.bigip._meta_data['icr_session']
        self.original_headers = copy.deepcopy(self.icr.session.headers)

    def __enter__(self):
        """Begins a new transaction context

        When a transaction begins, this method will automatically be called
        to set up the transaction.

        Transaction IDs are automatically retrieved for you and the
        appropriate headers are set so that operations in the Transaction
        Context will reference the transaction.

        Headers are preserved so that after you exit the transaction, you will
        be able to use the API object as you normally would.
        """

        self.transaction = self.transaction.create()

        self.icr.session.headers.update({
            'X-F5-REST-Coordination-Id': str(self.transaction.transId)
        })
        return self.bigip

    def __exit__(self, exc_type, exc_value, exc_tb):
        """Commit a transaction upon Context Manager exit

        Upon exit, the transaction will attempt to be committed. If the commit
        fails, the transaction will automatically be rolled back by the server
        performing the transaction.

        :param exc_type: The type of exception raised
        :param exc_value: Value of the exception raised
        :param exc_tb: Traceback
        NOTE: If the context exits without an exception, all three of the
        parameters will be None
        :returns: void
        """

        self.icr.session.headers = dict()
        if exc_tb is None:
            try:
                self.transaction.modify(state="VALIDATING",
                                        validateOnly=self.validate_only)
            except Exception as e:
                logging.debug(e)
                raise TransactionSubmitException(e)
            finally:
                self.icr.session.headers = self.original_headers
        self.icr.session.headers = self.original_headers



# coding=utf-8
#
# Copyright 2015-2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from f5.bigip.resource import UnnamedResource
from f5.sdk_exception import UnsupportedOperation


class Signature_Update(UnnamedResource):
    """BIG-IP® ASM Signature Update resource"""
    def __init__(self, asm):
        super(Signature_Update, self).__init__(asm)
        self._meta_data['object_has_stats'] = False
        self._meta_data['required_load_parameters'] = set()
        self._meta_data['required_json_kind'] =\
            'tm:asm:signature-update:sigfile-update-settingsstate'

    def update(self, **kwargs):
        """Update is not supported for Signature Update resource

                :raises: UnsupportedOperation
        """
        raise UnsupportedOperation(
            "%s does not support the update method" % self.__class__.__name__
        )



# coding=utf-8
#
# Copyright 2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from f5.bigiq.cm import Cm
from f5.bigiq.resource import PathElement
from f5.bigiq.shared import Shared
from f5.bigiq.tm import Tm
from icontrol.session import iControlRESTSession


class ManagementRoot(PathElement):
    """An interface to a single BIG-IP"""
    def __init__(self, hostname, username, password, **kwargs):
        timeout = kwargs.pop('timeout', 30)
        port = kwargs.pop('port', 443)
        icontrol_version = kwargs.pop('icontrol_version', '')

        # The BIG-IQ token is called "local", as opposed to BIG-IP's which
        # is called "tmos"
        auth_provider = kwargs.pop('auth_provider', 'local')
        verify = kwargs.pop('verify', False)
        if kwargs:
            raise TypeError('Unexpected **kwargs: %r' % kwargs)
        # _meta_data variable values
        iCRS = iControlRESTSession(
            username, password, timeout=timeout, auth_provider=auth_provider, verify=verify
        )
        # define _meta_data
        self._meta_data = {
            'allowed_lazy_attributes': [Cm, Shared, Tm],
            'hostname': hostname,
            'port': port,
            'uri': 'https://%s:%s/mgmt/' % (hostname, port),
            'icr_session': iCRS,
            'device_name': None,
            'local_ip': None,
            'bigip': self,
            'icontrol_version': icontrol_version,
            'username': username,
            'password': password,
            'tmos_version': None,
        }
        self._get_os_version()

    @property
    def hostname(self):
        return self._meta_data['hostname']

    @property
    def icontrol_version(self):
        return self._meta_data['icontrol_version']

    @property
    def tmos_version(self):
        return self._meta_data['tmos_version']

    def _get_os_version(self):
        connect = self._meta_data['bigip']._meta_data['icr_session']
        base_uri = self._meta_data['uri'] + \
            'shared/identified-devices/config/device-info'
        response = connect.get(base_uri)
        version = response.json()['version']
        self._meta_data['tmos_version'] = version

    @property
    def _debug(self):
        result = []
        if self.icrs._debug:
            result += self.icrs._debug
        return result



# coding=utf-8
#
# Copyright 2017 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""BIG-IP® system sflow module

REST URI
    ``http://localhost/mgmt/tm/sys/sflow``

GUI Path
   ``System->sFlow``

REST Kind
    ``tm:sys:sflow:*``
"""

# from distutils.version import LooseVersion
from f5.bigip.resource import Collection
from f5.bigip.resource import OrganizingCollection
from f5.bigip.resource import Resource
from f5.bigip.resource import UnnamedResource
# from f5.sdk_exception import UnsupportedMethod


class Sflow(OrganizingCollection):
    def __init__(self, sys):
        super(Sflow, self).__init__(sys)
        self._meta_data['allowed_lazy_attributes'] = [
            Global_Settings,
            Receivers]


class Global_Settings(OrganizingCollection):
    def __init__(self, sflow):
        super(Global_Settings, self).__init__(sflow)
        self._meta_data['allowed_lazy_attributes'] = [
            Http,
            Interface,
            System,
            Vlan]


class Http(UnnamedResource):
    def __init__(self, global_settings):
        super(Http, self).__init__(global_settings)
        self._meta_data['required_load_parameters'] = set()
        self._meta_data['required_json_kind'] =\
            'tm:sys:sflow:global-settings:http:httpstate'


class Interface(UnnamedResource):
    def __init__(self, global_settings):
        super(Interface, self).__init__(global_settings)
        self._meta_data['required_load_parameters'] = set()
        self._meta_data['required_json_kind'] =\
            'tm:sys:sflow:global-settings:interface:interfacestate'


class System(UnnamedResource):
    def __init__(self, global_settings):
        super(System, self).__init__(global_settings)
        self._meta_data['required_load_parameters'] = set()
        self._meta_data['required_json_kind'] =\
            'tm:sys:sflow:global-settings:system:systemstate'


class Vlan(UnnamedResource):
    def __init__(self, global_settings):
        super(Vlan, self).__init__(global_settings)
        self._meta_data['required_load_parameters'] = set()
        self._meta_data['required_json_kind'] =\
            'tm:sys:sflow:global-settings:vlan:vlanstate'


class Receivers(Collection):
    def __init__(self, sflow):
        super(Receivers, self).__init__(sflow)
        self._meta_data['allowed_lazy_attributes'] = [Receiver]
        self._meta_data['attribute_registry'] =\
            {'tm:sys:sflow:receiver:receiverstate': Receiver}


class Receiver(Resource):
    def __init__(self, receivers):
        super(Receiver, self).__init__(receivers)
        self._meta_data['required_creation_parameters'].update(
            ('name', 'address'))
        self._meta_data['required_json_kind'] =\
            'tm:sys:sflow:receiver:receiverstate'



# coding=utf-8
#
#  Copyright 2014-2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""BIG-IP® LTM auth submodule.

REST URI
    ``http://localhost/mgmt/tm/ltm/auth/``

GUI Path
    ``Local Traffic --> Profiles --> Authentication``

REST Kind
    ``tm:ltm:auth:*``
"""
from f5.bigip.resource import Collection
from f5.bigip.resource import OrganizingCollection
from f5.bigip.resource import Resource
from f5.sdk_exception import UnsupportedMethod


class Auth(OrganizingCollection):
    """BIG-IP® LTM Authentication organizing collection."""
    def __init__(self, ltm):
        super(Auth, self).__init__(ltm)
        self._meta_data['allowed_lazy_attributes'] = [
            Crldp_Servers,
            Kerberos_Delegations,
            Ldaps,
            Ocsp_Responders,
            Profiles,
            Radius_s,
            Radius_Servers,
            Ssl_Cc_Ldaps,
            Ssl_Crldps,
            Ssl_Ocsps,
            Tacacs_s
        ]


class Crldp_Servers(Collection):
    """BIG-IP® LTM Auth Crldp Server collection"""
    def __init__(self, ltm):
        super(Crldp_Servers, self).__init__(ltm)
        self._meta_data['allowed_lazy_attributes'] = [Crldp_Server]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:auth:crldp-server:crldp-serverstate': Crldp_Server}


class Crldp_Server(Resource):
    def __init__(self, crldp_servers):
        """BIG-IP® LTM Auth Crldp Server resource"""
        super(Crldp_Server, self).__init__(crldp_servers)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:auth:crldp-server:crldp-serverstate'
        self._meta_data['required_creation_parameters'].update(('host',))


class Kerberos_Delegations(Collection):
    """BIG-IP® LTM Auth Kerberos Delegation collection"""
    def __init__(self, ltm):
        super(Kerberos_Delegations, self).__init__(ltm)
        self._meta_data['allowed_lazy_attributes'] = [Kerberos_Delegation]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:auth:kerberos-delegation:kerberos-delegationstate':
                Kerberos_Delegation}


class Kerberos_Delegation(Resource):
    """BIG-IP® LTM Auth Kerberos Delegation resource"""
    def __init__(self, kerberos_delegations):
        super(Kerberos_Delegation, self).__init__(kerberos_delegations)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:auth:kerberos-delegation:kerberos-delegationstate'
        self._meta_data['required_creation_parameters'].update(
            ('serverPrincipal', 'clientPrincipal',))


class Ldaps(Collection):
    """BIG-IP® LTM Auth Ldap collection"""
    def __init__(self, ltm):
        super(Ldaps, self).__init__(ltm)
        self._meta_data['allowed_lazy_attributes'] = [Ldap]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:auth:ldap:ldapstate': Ldap}


class Ldap(Resource):
    """BIG-IP® LTM Auth Ldap resource"""
    def __init__(self, ldaps):
        super(Ldap, self).__init__(ldaps)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:auth:ldap:ldapstate'
        self._meta_data['required_creation_parameters'].update(('servers',))


class Ocsp_Responders(Collection):
    """BIG-IP® LTM Auth Ocsp Responder collection"""
    def __init__(self, ltm):
        super(Ocsp_Responders, self).__init__(ltm)
        self._meta_data['allowed_lazy_attributes'] = [Ocsp_Responder]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:auth:ocsp-responder:ocsp-responderstate': Ocsp_Responder}


class Ocsp_Responder(Resource):
    """BIG-IP® LTM Auth Ocsp Responder resource"""
    def __init__(self, ocsp_responders):
        super(Ocsp_Responder, self).__init__(ocsp_responders)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:auth:ocsp-responder:ocsp-responderstate'


class Profiles(Collection):
    """BIG-IP® LTM Auth Profile collection"""
    def __init__(self, ltm):
        super(Profiles, self).__init__(ltm)
        self._meta_data['allowed_lazy_attributes'] = [Profile]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:auth:profile:profilestate': Profile}


class Profile(Resource):
    """BIG-IP® LTM Auth Profile resource"""
    def __init__(self, profiles):
        super(Profile, self).__init__(profiles)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:auth:profile:profilestate'
        self._meta_data['required_creation_parameters'].update(
            ('defaultsFrom', 'configuration',))

    def update(self, **kwargs):
        '''Update is not supported for LTM Auth Profiles

        :raises: UnsupportedOperation
        '''
        raise UnsupportedMethod(
            "%s does not support the modify method" % self.__class__.__name__
        )


class Radius_s(Collection):
    """BIG-IP® LTM Auth Radius collection"""
    def __init__(self, ltm):
        super(Radius_s, self).__init__(ltm)
        self._meta_data['allowed_lazy_attributes'] = [Radius]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:auth:radius:radiusstate': Radius}


class Radius(Resource):
    """BIG-IP® LTM Auth Radius resource"""
    def __init__(self, radius_s):
        super(Radius, self).__init__(radius_s)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:auth:radius:radiusstate'


class Radius_Servers(Collection):
    """BIG-IP® LTM Auth Radius Server collection"""
    def __init__(self, ltm):
        super(Radius_Servers, self).__init__(ltm)
        self._meta_data['allowed_lazy_attributes'] = [Radius_Server]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:auth:radius-server:radius-serverstate': Radius_Server}


class Radius_Server(Resource):
    """BIG-IP® LTM Auth Radius Server resource"""
    def __init__(self, radius_server_s):
        super(Radius_Server, self).__init__(radius_server_s)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:auth:radius-server:radius-serverstate'
        self._meta_data['required_creation_parameters'].update(
            ('secret', 'server',))


class Ssl_Cc_Ldaps(Collection):
    """BIG-IP® LTM Auth SSL CC LDAP collection"""
    def __init__(self, ltm):
        super(Ssl_Cc_Ldaps, self).__init__(ltm)
        self._meta_data['allowed_lazy_attributes'] = [Ssl_Cc_Ldap]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:auth:ssl-cc-ldap:ssl-cc-ldapstate': Ssl_Cc_Ldap}


class Ssl_Cc_Ldap(Resource):
    """BIG-IP® LTM Auth SSL CC LDAP resource"""
    def __init__(self, ssl_cc_ldaps):
        super(Ssl_Cc_Ldap, self).__init__(ssl_cc_ldaps)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:auth:ssl-cc-ldap:ssl-cc-ldapstate'
        self._meta_data['required_creation_parameters'].update(
            ('servers', 'userKey',))


class Ssl_Crldps(Collection):
    """BIG-IP® LTM Auth SSL CLRDP collection"""
    def __init__(self, ltm):
        super(Ssl_Crldps, self).__init__(ltm)
        self._meta_data['allowed_lazy_attributes'] = [Ssl_Crldp]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:auth:ssl-crldp:ssl-crldpstate': Ssl_Crldp}


class Ssl_Crldp(Resource):
    """BIG-IP® LTM Auth SSL CLRDP resource"""
    def __init__(self, ssl_crldps):
        super(Ssl_Crldp, self).__init__(ssl_crldps)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:auth:ssl-crldp:ssl-crldpstate'


class Ssl_Ocsps(Collection):
    """BIG-IP® LTM Auth SSL OCSP collection"""
    def __init__(self, ltm):
        super(Ssl_Ocsps, self).__init__(ltm)
        self._meta_data['allowed_lazy_attributes'] = [Ssl_Ocsp]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:auth:ssl-ocsp:ssl-ocspstate': Ssl_Ocsp}


class Ssl_Ocsp(Resource):
    """BIG-IP® LTM Auth SSL OCSP resource"""
    def __init__(self, ssl_ocsps):
        super(Ssl_Ocsp, self).__init__(ssl_ocsps)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:auth:ssl-ocsp:ssl-ocspstate'


class Tacacs_s(Collection):
    """BIG-IP® LTM Auth TACACS+ Server collection"""
    def __init__(self, ltm):
        super(Tacacs_s, self).__init__(ltm)
        self._meta_data['allowed_lazy_attributes'] = [Tacacs]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:auth:tacacs:tacacsstate': Tacacs}


class Tacacs(Resource):
    """BIG-IP® LTM Auth TACACS+ Server resource"""
    def __init__(self, tacacs_s):
        super(Tacacs, self).__init__(tacacs_s)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:auth:tacacs:tacacsstate'
        self._meta_data['required_creation_parameters'].update(
            ('secret', 'servers', 'service'))



#!/usr/bin/env python
# Copyright 2017 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from __future__ import print_function

import errno
import getopt
import glob
import json
import os
import re
import sys

from collections import deque
from collections import namedtuple

try: # for pip >= 10
    from pip._internal.req import parse_requirements as p_reqs
except ImportError: # for pip <= 9.0.3
    from pip.req import parse_requirements as p_reqs


def construct_cfgs(**kargs):
    """construct_cfgs

Performs actions to construct either the setup.cfg (rpm) or stdeb.cfg (deb)
files as per the operating system specified.  This construction is done as
per the setup_requirements.txt file from within the working directory
specified.

This is a very tempermental function by design.  It intentionally exits
with a non-zero if/when an error has occurred on its own.  Therefore, it is
not suggested to use this function if you intend to get control back again.
    """
    docker_dir = _check_args(**kargs)
    if not docker_dir or not os.path.isdir(docker_dir):
        print("Unable to determine the %s/%s combo under supported versioning"
              % (kargs['operating_system'], kargs['version']))
        exit_cleanly(error_number=errno.ENOSYS)
    if kargs['operating_system'] == 'redhat':
        _build_setup_cfg(kargs['working_directory'])
    elif kargs['operating_system'] == 'ubuntu':
        _build_stdeb_cfg(kargs['working_directory'])
    else:
        print("Unsure of what to do... operating_system(%s) is not recognized!"
              % kargs['operating_system'])


def _construct_cfgs_from_json(args):
    Args = namedtuple('Args', 'setup, reqs, fmt, start')
    rpm = Args(args['setup_cfg'], args['setup_requirements'], '%s%s',
               'requires = ')
    deb = Args(args['stdeb_cfg'], args['setup_requirements'], '%s (%s), ',
               'Depends:\n     ')
    _construct_file(rpm.setup, rpm.reqs, rpm.fmt, rpm.start)
    _construct_file(deb.setup, deb.reqs, deb.fmt, deb.start)
    print("Successfully constructed %s and\n%s" %
          (args['setup_cfg'], args['stdeb_cfg']))


def _read_in_cfgs(cfg):
    cfgs = deque()
    break_re = re.compile('^[^#\W]')
    req_re = re.compile('^Depends:|^requires\s*=\s*')
    try:
        with open(cfg, 'r') as fh:
            read = True
            line = fh.readline()
            while line:
                read = True if break_re.search(line) and not read else read
                if req_re.search(line):
                    read = False
                if read:
                    cfgs.append(line)
                line = fh.readline()
    except IOError:
        pass  # we then have empty contents... all well...
    return cfgs


def _build_setup_cfg(wkg):
    setup_cfg = wkg + "/setup.cfg"
    setup = wkg + "/setup_requirements.txt"
    fmt = "%s%s"
    _construct_file(setup_cfg, setup, fmt, 'requires = ')


def _build_stdeb_cfg(wkg):
    setup_cfg = glob.glob(wkg + "/*-dist/deb_dist/stdeb.cfg")
    setup = wkg + "/setup_requirements.txt"
    fmt = "%s (%s),"
    if not setup_cfg:
        dist = glob.glob(wkg + "/*-dist")
        if not dist:
            print(str(EnvironmentError("Unable to find a *-dist directory")))
            print("No dist directory under: " + wkg)
            exit_cleanly(error_number=errno.ENOSYS)
        deb_dist = dist[0] + "/deb_dist"
        try:
            os.mkdir(deb_dist)
        except IOError as Error:
            if not os.path.isdir(deb_dist):
                print(str(Error))
                print("Unable to determine the existence of: " + deb_dist)
                exit_cleanly(error_number=errno.ENOSYS)
        setup_cfg = deb_dist + "/stdeb.cfg"
    else:
        setup_cfg = setup_cfg[0]
    _construct_file(setup_cfg, setup, fmt, 'Depends:\n    ')


def _construct_file(setup_cfg, setup, fmt, start):
    if not os.path.isfile(setup) or not os.access(setup, os.R_OK):
        print(setup + " does not exist or is not readable")
        exit_cleanly(error_number=errno.ENOSYS)
    contents = _read_in_cfgs(setup_cfg) if os.path.isfile(setup_cfg) else deque()
    parsed_reqs = list(map(lambda x: (x.req), p_reqs(setup, session="pkg")))
    if not parsed_reqs:
        print("Nothing to do!\n%s\nDoes not contain any reqs parsable!" % setup)
        exit_cleanly(error_number=0)
    try:
        with open(setup_cfg, 'w') as fh:
            if contents:
                while contents:
                    fh.write(contents.popleft())
            fh.write(start)
            for count in range(len(parsed_reqs)):
                req = parsed_reqs[count]
                if 'Depends' in start:
                    # special case for debian...
                    name = str(req.name) if 'python-' in str(req.name) else 'python-' + str(req.name)
                else:
                    name = str(req.name)
                fh.write(fmt % (name, str(req.specifier)))
                if count != len(parsed_reqs) - 1:
                    fh.write('\n    ')
    except IOError as Error:
        print(Error)
        exit_cleanly(error_number=errno.EIO)


def _check_args(operating_system=None, version=None, working_directory=None):
    accepted_docker_dir = None
    try:
        docker_dir = str(working_directory) + "/*-dist/Docker/%s/%s"
    except TypeError as Error:
        print("working_directory is of invalid type! (%s)" %
              str(type(working_directory)))
        exit_cleanly(error_number=errno.EINVAL)
    except Exception as Error:
        print(str(Error))
        print("'%s' is not a valid working_directory!" %
              str(working_directory))
        exit_cleanly(error_number=errno.EINVAL)
    if operating_system not in ['ubuntu', 'redhat']:
        print("'%s' is not a valid or recognized operating_system option" %
              operating_system)
        exit_cleanly(error_number=errno.EINVAL)
    else:
        possible = glob.glob(docker_dir % (operating_system, version))
        if possible:
            accepted_docker_dir = possible[0]
    if not accepted_docker_dir:
        print("(operating_system=%s, version=%s, working_directory=%s)" %
              (operating_system, version, working_directory))
        print("Are not acceptable arguments resulting in a Docker " +
              "file location")
        exit_cleanly()
    return accepted_docker_dir


def exit_cleanly(error_number=0):
    """exit_cleanly

Performs standard error notification and exiting statements as necessary.  This
assures more consistent error handling within the script.
    """
    default = "An Unknown error has occurred!"
    descriptions = \
        {22: 'An improper input error has occurred.  Please see above stmt(s)',
         29: 'An operation failed.  Please see above stmt(s)',
         5: 'An IO Error has occurred.  Pelase see above stmt(s)'}
    try:
        error_number = int(error_number)
    except TypeError:
        stmt = default
        error_number = -1
    if error_number == 0:
        stmt = "No error has been detected!"
    elif not isinstance(error_number, int) and hasattr(errno, error_number):
        error_number = getattr(errno, error_number)
        stmt = descriptions[error_number] if error_number in descriptions \
            else default
    elif error_number in descriptions:
        stmt = descriptions[error_number]
    else:
        stmt = default
    if error_number:
        print("""
%s [--opt [option]]
    With opts:
        working_directory - the full path to the working directory
        operating_system - the full name of the operating system lower case
        version - the version of the operating system

    All of these options must be supplied, and if one is missing or if there is
    no corresponding:
        <working_directory>/*-dist/Docker/<operating_system>/<version>
    Directory, then this script will exit cleanly reporting it as an error.
            """ % sys.argv[0])
    print("(%d) %s" % (error_number, stmt), file=sys.stderr)
    sys.exit(error_number)


def _load_json(json_fl):
    try:
        with open(json_fl, 'r') as fh:
            data = json.loads(fh.read())
    except Exception as Error:
        print(Error)
        print("Unable to read in " + json_fl)
    return data


def get_args():
    """get_args

This function extracts the script arguments within the arguments variable and
interprets their meaning before returning such content.
    """
    expected = ['working_directory', 'operating_system', 'version']
    possible = ['json'].extend(expected)
    try:
        opts, args = getopt.getopt(sys.argv[1:], '', map(lambda x: ('%s=' % x),
                                   possible))
    except getopt.GetoptError as err:
        print(str(err))
        exit_cleanly(error_number=errno.EINVAL)
    arguments = dict()
    for o, a in opts:
        option = re.sub('^-*', '', o)
        if 'json' in option:
            arguments = _load_json(a)
            break
        if option in expected:
            arguments[option] = a
    error = 0
    for item in expected:
        if item not in arguments:
            print("Missing: %s from arguments" % item)
            error = errno.EINVAL
    if error:
        exit_cleanly(error_number=error)
    return arguments


def main():
    """main

The entrypoint function.  This function should also handle any runtime
errors and exceptions in a cleanly fashon.
    """
    try:
        args = get_args()
        if 'setup_cfg' in args and 'stdeb_cfg' in args:
            _construct_cfgs_from_json(args)
        else:
            construct_cfgs(**args)
    except Exception as Error:
        print(str(Error))
        print("Exiting cleanly...")
        exit_cleanly()
    exit_cleanly(error_number=0)


if __name__ == '__main__':
    main()



# coding=utf-8
#
#  Copyright 2014-2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""BIG-IP® Network interface module.

REST URI
    ``http://localhost/mgmt/tm/net/interface``

GUI Path
    ``Network --> Interfaces``

REST Kind
    ``tm:net:interface:*``
"""

from f5.bigip.mixins import ExclusiveAttributesMixin
from f5.bigip.resource import Collection
from f5.bigip.resource import Resource
from f5.sdk_exception import UnsupportedOperation


class Interfaces(Collection):
    """BIG-IP® network interface collection"""
    def __init__(self, net):
        super(Interfaces, self).__init__(net)
        self._meta_data['allowed_lazy_attributes'] = [Interface]
        self._meta_data['attribute_registry'] = {
            'tm:net:interface:interfacestate': Interface
        }


class Interface(Resource, ExclusiveAttributesMixin):
    """BIG-IP® network interface collection"""
    def __init__(self, interface_s):
        super(Interface, self).__init__(interface_s)
        self._meta_data['required_json_kind'] =\
            'tm:net:interface:interfacestate'
        self._meta_data['exclusive_attributes'].append(('enabled', 'disabled'))

    def create(self, **kwargs):
        """Create is not supported for interfaces.

        :raises: :exc:`~f5.BIG-IP.resource.UnsupportedOperation`
        """
        raise UnsupportedOperation(
            "BigIP interfaces cannot be created by users")

    def delete(self):
        """Delete is not supported for interfaces.

        :raises: :exc:`~f5.BIG-IP.resource.UnsupportedOperation`
        """
        raise UnsupportedOperation(
            "BigIP interfaces cannot be deleted by users")



# coding=utf-8
#
# Copyright 2017 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""BIG-IP® Network trunk module.

REST URI
    ``http://localhost/mgmt/tm/net/trunk``

GUI Path
    ``Network --> Trunks``

REST Kind
    ``tm:net:trunk:*``
"""

from f5.bigip.mixins import ExclusiveAttributesMixin
from f5.bigip.resource import Collection
from f5.bigip.resource import Resource


class Trunks(Collection):
    """BIG-IP® network route collection"""
    def __init__(self, net):
        super(Trunks, self).__init__(net)
        self._meta_data['allowed_lazy_attributes'] = [Trunk]
        self._meta_data['attribute_registry'] = {
            'tm:net:trunk:trunkstate': Trunk
        }


class Trunk(Resource, ExclusiveAttributesMixin):
    def __init__(self, trunks):
        super(Trunk, self).__init__(trunks)
        self._meta_data['required_json_kind'] = 'tm:net:trunk:trunkstate'
        self._meta_data['read_only_attributes'].append(
            ('network', 'cfgMbrCount', 'workingMbrCount', 'macAddress',
             'bandwidth')
        )



#!/usr/bin/env python
# Copyright 2017 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import errno
import glob
import json
import os
import re
import sys


def add_package(pkg_type, pkg_name, working=None):
    """add_package

Adds a package to the existing config.JSON file.  This is a standalone function
to handle this functionality.  The existing config.JSON is read in and is left
unchanged.
    """
    kvp = {pkg_type: pkg_name}
    working = os.getcwd() if not working else os.path.abspath(working)
    if '-dist/scripts' in working:
        config = working + "/config.JSON"
    elif '-dist' in working:
        config = working + "/scripts/config.JSON"
    else:
        config = working + "/*-dist/scripts/config.JSON"
        entropy = glob.glob(config)
        if entropy:
            config = entropy[0]  # we'll just assume it's the first one...
    print("config", config, 'working', working)
    config = load_config(config)
    config.update(kvp)
    export_to_json(config)


def exit_cleanly(errnum=None):
    """exit_cleanly

Exits the runtime cleanly using SystemExit.  This is the official, handling
exception here as this is mostly meant to be a standalone script.
    """
    default = "An Unknown Error has Occurred!"
    cases = {errno.EINVAL: "Improper or invalid input value",
             errno.ESPIPE: "Could not complete action due to a broken" +
             " dependency",
             errno.ENOSYS: "Could not complete action due to unexpected setup",
             errno.EIO: "Could not access an expected file",
             errno.EPERM: "Could not access an item due to permissions issues",
             -1: default}
    help_stmt = """
%s [--opt [option]]
    With opts:
        working_directory - ONLY use this if you're overwriting `pwd`!
    """ % (sys.argv[0])
    if not errnum:
        errnum = 0
    elif not isinstance(errnum, int) and hasattr(errno, errnum):
        errnum = getattr(errno, errnum)
    try:
        errnum = int(errnum)
    except TypeError:
        errnum = -1
    if errnum == 0:
        help_stmt = ''
        stmt = "Successful in configuration!"
    elif errnum in cases:
        stmt = cases[errnum]
    else:
        stmt = default
    print("%s\n\n%s" % (stmt, help_stmt))


def get_env(working_directory=None):
    """get_env

This function grabs key/value pair items and assigns them for the
config.JSON.  This essentially checks the environment for key locations within
the filesystem.
    """
    working = working_directory if working_directory else os.getcwd()
    dist_dirs = glob.glob(working + "/f5-*-dist")
    print(dist_dirs)
    dist_dir_re = re.compile('/([^/]+)-dist/?')
    if dist_dirs:
        dist_dir = dist_dirs[0]
        match = dist_dir_re.search(dist_dir)
        if match:
            project_name = match.group(1)
        else:
            print("Unrecognized run location:\n" + working)
            exit_cleanly(errnum=errno.EIO)
    elif '-dist/scripts' in working:
        match = dist_dir_re.search(working)
        if match:
            project_name = match.group(1)
    else:
        print("Unable to determine the *-dist directory from " + working)
        exit_cleanly(errnum=errno.ENOSYS)
    stdeb_cfg = dist_dir + "/deb_dist/stdeb.cfg"
    setup_cfg = working + "/setup.cfg"
    stp_reqs = working + "/setup_requirements.txt"
    scripts = dist_dir + "/scripts"
    env = {'working': working, 'dist_dir': dist_dir, 'stdeb_cfg': stdeb_cfg,
           'setup_requirements': stp_reqs, 'setup_cfg': setup_cfg,
           'scripts': scripts, 'project': project_name}
    return env


def export_to_json(env):
    """export_to_json

This function takes in a dictionary object and stores it within the config.JSON
file.
    """
    json_fl = env['scripts'] + "/config.JSON"
    with open(json_fl, 'w') as fh:
        fh.write(json.dumps(env, sort_keys=True, indent=4,
                            separators=(',', ': ')))


def load_config(filename):
    """load_config

This function will take in a file location on the file system and extract the
JSON content found within and return a python-interpretted form of the object.
    """
    config = None
    if os.path.isfile(filename) and os.access(filename, os.R_OK):
        with open(filename, 'r') as fh:
            config = json.loads(fh.read())
    return config


def main():
    """main

This is the entrypoint for this script when executed on its own as a standalone
script.
    """
    env = get_env()
    export_to_json(env)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# Copyright 2017 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os
import subprocess

from collections import namedtuple

# used for terminal:
Result = namedtuple('Result', 'succeeded, cli, error')


class Hold(object):
    """Hold

A basic text caching object for later use of printing into the terminal,
storing into a return object, or mimicking the standard IO commands of a file
object.

This object is handy for reducing code size, but offering a transparent
functionality to a user using the terminal function.
    """
    def __init__(self):
        self.held = str()

    def __str__(self):
        return self.get_held()

    def __enter__(self):
        return self

    def __exit__(self):
        self.held = ''

    def writeline(self, addition):
        """writeline()

Functions like a file.writeline() call; however, it stores into the object's
cached memory rather than a file's IO.
        """
        addition = addition.strip()
        self.held = self.held + addition + "\n"

    def write(self, addition):
        """write()

Functions like a file.write() call; however, it stores into the object's cached
memory rather than a file's IO.
        """
        self.held = self.held + addition

    def output(self):
        """output()

Dumps the contents of the cache as a string into the terminal.
        """
        print(self.held)

    def get_held(self):
        """get_held()

Returns the cached contents as a str() object.
        """
        return self.held


def terminal(cmd, log=None, args=[], want_result=True, multi_stmt=False):
    """terminal

Usage:
    terminal(cmd, log=None, args=[], want_result=True, multi_stmt=False)

    Args:
        cmd - list or tuple(list) that contains the command with args
            commonly delimited by spaces
    Opts:
        log - string containing the log location where to store the CLI
            stdout result from running the command on the terminal
        args - list of arguments within the command statement that would
            normally be delimited by spaces.  Note that if the cmd argument
            is already a list, then it will be extended by this list.  If it
            is a str, then it will be put into a list with this list
            extended after it.
        want_result - boolean contaiining whether or not the result,
            namedtuple will be returned
        multi_stmt - boolean enabling there to be a multi-command stmt.  If
            this is left as False, then cmd statements in the arguments or
            as a part of the cmd argument will be ignored and not
            interpretted, or the OSError will be thrown over the fact that
            the cmd will not be allowed.
    Returns:
        result - A named tuple containing the following flags:
            succeeded - boolean on whether or not the command succeeded
            cli - the cli's stdout
            error - the exception that was thrown if one was over whether or
                not the command failed at the CLI

This function will attempt to execute a CLI command at the /bin/bash
terminal.
    """
    Error = None
    succeeded = False
    if log and isinstance(log, str) and os.access(log, os.W_OK):
        fh = open(log, 'w')
    else:
        fh = Hold()
    if args and isinstance(cmd, str):
        cmd = [cmd]
        cmd.extend(args)
    try:
        proc = subprocess.Popen(cmd, shell=multi_stmt, stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE)
        # fh.write(subprocess.check_output(cmd, shell=multi_stmt))
        out, err = proc.communicate()
        fh.write('Out:\n' + str(out))
        fh.write('Err:\n' + str(err))
        succeeded = True if proc.returncode == 0 else False
    except OSError as Error:
        fh.write(str(Error))
    except subprocess.CalledProcessError as Error:
        fh.write(str(Error))

    try:
        fh.close()
        result = Result(succeeded, None, Error)
    except Exception:
        if want_result:
            result = Result(succeeded, fh.get_held(), Error)
        elif not want_result:
            result = Result(succeeded, None, Error)
            fh.output()
    return result



# coding=utf-8
#
# Copyright 2017 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""BIG-IP® auth module

REST URI
    ``http://localhost/mgmt/tm/auth/remote-role``

GUI Path
    ``System --> Users --> Remote Role Groups``

REST Kind
    ``tm:auth:remote-role:*``
"""

from f5.bigip.resource import Collection
from f5.bigip.resource import Resource
from f5.bigip.resource import UnnamedResource
from f5.sdk_exception import UnsupportedOperation


class Remote_Role(UnnamedResource):
    """BIG-IP® auth remote role resource"""
    def __init__(self, auth):
        super(Remote_Role, self).__init__(auth)
        self._meta_data['allowed_lazy_attributes'] = [Role_Infos]
        self._meta_data['required_json_kind'] = \
            'tm:auth:remote-role:remote-rolestate'
        self._meta_data['attribute_registry'] = \
            {'tm:auth:remote-role:role-info:role-infocollectionstate': Role_Infos}


class Role_Infos(Collection):
    """BIG-IP® remote role role-info collection"""
    def __init__(self, auth):
        super(Role_Infos, self).__init__(auth)
        self._meta_data['allowed_lazy_attributes'] = [Role_Info]
        self._meta_data['required_json_kind'] = \
            'tm:auth:remote-role:role-info:role-infocollectionstate'
        self._meta_data['attribute_registry'] = \
            {'tm:auth:remote-role:role-info:role-infostate': Role_Info}


class Role_Info(Resource):
    """BIG-IP® remote role role-info resource"""
    def __init__(self, role_infos):
        super(Role_Info, self).__init__(role_infos)
        self._meta_data['required_json_kind'] = \
            'tm:auth:remote-role:role-info:role-infostate'

    def update(self, **kwargs):
        '''Update is not supported for Auth Remote Role - Role Info Objects

        :raises: UnsupportedOperation
        '''
        raise UnsupportedOperation(
            "%s does not support update, use modify instead" % self.__class__.__name__
        )



# coding=utf-8
#
#  Copyright 2014-2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""BIG-IP® LTM ifile submodule.

REST URI
    ``http://localhost/mgmt/tm/ltm/ifile/``

GUI Path
    ``Local Traffic --> iRules --> iFiles``

REST Kind
    ``tm:ltm:ifile*``
"""

from f5.bigip.resource import Collection
from f5.bigip.resource import Resource


class Ifiles(Collection):
    """BIG-IP® LTM iFiles collection"""
    def __init__(self, ltm):
        super(Ifiles, self).__init__(ltm)
        self._meta_data['allowed_lazy_attributes'] = [Ifile]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:ifile:ifilestate': Ifile}


class Ifile(Resource):
    """BIG-IP® LTM iFiles resource"""
    def __init__(self, ifile_s):
        super(Ifile, self).__init__(ifile_s)
        self._meta_data['required_json_kind'] = 'tm:ltm:ifile:ifilestate'
        self._meta_data['required_creation_parameters'].update(
            ('name', 'fileName')
        )



# coding=utf-8
#
# Copyright 2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""BIG-IP® system failover module

REST URI
    ``http://localhost/mgmt/tm/sys/failover``

GUI Path
    ``System --> Failover``

REST Kind
    ``tm:sys:failover:*``
"""

from f5.bigip.mixins import CommandExecutionMixin
from f5.bigip.resource import UnnamedResource


class Failover(UnnamedResource, CommandExecutionMixin):
    '''BIG-IP® Failover stats and state change.

    The failover object only supports load, update, and refresh because it is
     an unnamed resource.

    To force the unit to standby call the ``update()`` method as follows:

    .. code-block:: python
        f.update(command='run', standby=None, trafficGroup='mytrafficgroup')

    .. note::

        This is an unnamed resource so it has not ~Partition~Name pattern
        at the end of its URI.
    '''
    def __init__(self, sys):
        super(Failover, self).__init__(sys)
        self._meta_data['required_load_parameters'] = set()
        self._meta_data['required_json_kind'] =\
            'tm:sys:failover:failoverstats'
        self._meta_data['allowed_commands'].append('run')

    def update(self, **kwargs):
        """Update is not supported for Failover

        :raises: UnsupportedOperation
        """
        raise self.UnsupportedMethod(
            "%s does not support the update method" % self.__class__.__name__
        )

    def exec_cmd(self, command, **kwargs):
        """Defining custom method to append 'exclusive_attributes'.

        WARNING: Some parameters are hyphenated therefore the function
                 will need to utilize variable keyword argument syntax.
                 This only applies when utilCmdArgs method is not in use.

                 eg.

                 param_set ={'standby':True 'traffic-group': 'traffic-group-1'}
                 bigip.tm.sys.failover.exec_cmd('run', **param_set

        The 'standby' attribute cannot be present with either 'offline'
        or 'online' attribute, whichever is present. Additionally
        we check for existence of same attribute values in
        'offline' and 'online' if both present.


        note:: There is also another way of using failover endpoint,
               by the means of 'utilCmdArgs' attribute, here the syntax
               will resemble more that of the 'tmsh run sys failover...'
               command.
                    eg. exec_cmd('run', utilCmdArgs='standby traffic-group
                    traffic-group-1')

        :: raises InvalidParameterValue
        """

        kwargs = self._reduce_boolean_pair(kwargs, 'online', 'offline')
        if 'offline' in kwargs:
            self._meta_data['exclusive_attributes'].append(
                ('offline', 'standby'))

        if 'online' in kwargs:
            self._meta_data['exclusive_attributes'].append(
                ('online', 'standby'))
        self._is_allowed_command(command)
        self._check_command_parameters(**kwargs)
        return self._exec_cmd(command, **kwargs)

    def toggle_standby(self, **kwargs):
        """Toggle the standby status of a traffic group.

         WARNING: This method which used POST obtains json keys from the device
         that are not available in the response to a GET against the same URI.

         NOTE: This method method is deprecated and probably will be removed,
               usage of exec_cmd is encouraged.
        """

        trafficgroup = kwargs.pop('trafficgroup')
        state = kwargs.pop('state')
        if kwargs:
            raise TypeError('Unexpected **kwargs: %r' % kwargs)
        arguments = {'standby': state, 'traffic-group': trafficgroup}
        return self.exec_cmd('run', **arguments)



# Copyright 2016 F5 Networks Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#

'''These exceptions for clustering, devicegroup, and trustdomain.'''

from f5.sdk_exception import F5SDKError


class ClusterError(F5SDKError):
    def __init__(self, *args, **kwargs):
        super(ClusterError, self).__init__(*args, **kwargs)


class DeviceGroupError(F5SDKError):
    pass


class TrustDomainError(F5SDKError):
    pass


class AlreadyManagingCluster(ClusterError):
    pass


class ClusterNotSupported(ClusterError):
    pass


class ClusterOperationNotSupported(ClusterError):
    pass


class DeviceAlreadyInTrustDomain(TrustDomainError):
    pass


class MissingRequiredDeviceGroupParameter(ClusterError):
    pass


class NoClusterToManage(ClusterError):
    pass


class DeviceGroupNotSupported(DeviceGroupError):
    pass


class DeviceGroupOperationNotSupported(DeviceGroupError):
    pass


class DeviceNotTrusted(TrustDomainError):
    pass


class TrusteeNotInTrustDomain(TrustDomainError):
    pass


class UnexpectedDeviceGroupState(DeviceGroupError):
    pass


class UnexpectedDeviceGroupDevices(DeviceGroupError):
    pass


class UnexpectedDeviceGroupType(DeviceGroupError):
    pass



# coding=utf-8
#
# Copyright 2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""BIG-IP® system file module

REST URI
    ``http://localhost/mgmt/tm/sys/provision``

GUI Path
    N/A

REST Kind
    ``tm:sys:provision:*``
"""

from f5.bigip.resource import OrganizingCollection
from f5.bigip.resource import UnnamedResource


class Provision(OrganizingCollection):
    def __init__(self, sys):
        super(Provision, self).__init__(sys)
        self._meta_data['allowed_lazy_attributes'] = [
            Afm,
            Am,
            Apm,
            Asm,
            Avr,
            Dos,
            Fps,
            Gtm,
            Ili,
            Ilx,
            Lc,
            Ltm,
            Pem,
            Swg,
            Urldb,
            Vcmp
        ]


class Afm(UnnamedResource):
    """BIG-IP® system provision afm resource

    The afm object only supports load and update because it is an
    unnamed resource.

    .. note::

        This is an unnamed resource so it has not ~Partition~Name pattern
        at the end of its URI.
    """
    def __init__(self, provision):
        super(Afm, self).__init__(provision)
        self._meta_data['required_load_parameters'] = set()
        self._meta_data['required_json_kind'] =\
            'tm:sys:provision:provisionstate'


class Am(UnnamedResource):
    """BIG-IP® system provision afm resource

    The am object only supports load and update because it is an
    unnamed resource.

    .. note::

        This is an unnamed resource so it has not ~Partition~Name pattern
        at the end of its URI.
    """
    def __init__(self, provision):
        super(Am, self).__init__(provision)
        self._meta_data['required_load_parameters'] = set()
        self._meta_data['required_json_kind'] =\
            'tm:sys:provision:provisionstate'


class Apm(UnnamedResource):
    """BIG-IP® system provision apm resource

    The afm object only supports load and update because it is an
    unnamed resource.

    .. note::

        This is an unnamed resource so it has not ~Partition~Name pattern
        at the end of its URI.
    """
    def __init__(self, provision):
        super(Apm, self).__init__(provision)
        self._meta_data['required_load_parameters'] = set()
        self._meta_data['required_json_kind'] =\
            'tm:sys:provision:provisionstate'


class Asm(UnnamedResource):
    """BIG-IP® system provision asm resource

    The afm object only supports load and update because it is an
    unnamed resource.

    .. note::

        This is an unnamed resource so it has not ~Partition~Name pattern
        at the end of its URI.
    """
    def __init__(self, provision):
        super(Asm, self).__init__(provision)
        self._meta_data['required_load_parameters'] = set()
        self._meta_data['required_json_kind'] =\
            'tm:sys:provision:provisionstate'


class Avr(UnnamedResource):
    """BIG-IP® system provision avr resource

    The afm object only supports load and update because it is an
    unnamed resource.

    .. note::

        This is an unnamed resource so it has not ~Partition~Name pattern
        at the end of its URI.
    """
    def __init__(self, provision):
        super(Avr, self).__init__(provision)
        self._meta_data['required_load_parameters'] = set()
        self._meta_data['required_json_kind'] =\
            'tm:sys:provision:provisionstate'


class Dos(UnnamedResource):
    """BIG-IP® system provision dos resource

    The afm object only supports load and update because it is an
    unnamed resource.

    .. note::

        This is an unnamed resource so it has not ~Partition~Name pattern
        at the end of its URI.
    """
    def __init__(self, provision):
        super(Dos, self).__init__(provision)
        self._meta_data['required_load_parameters'] = set()
        self._meta_data['required_json_kind'] =\
            'tm:sys:provision:provisionstate'


class Fps(UnnamedResource):
    """BIG-IP® system provision fps resource

    The afm object only supports load and update because it is an
    unnamed resource.

    .. note::

        This is an unnamed resource so it has not ~Partition~Name pattern
        at the end of its URI.
    """
    def __init__(self, provision):
        super(Fps, self).__init__(provision)
        self._meta_data['required_load_parameters'] = set()
        self._meta_data['required_json_kind'] =\
            'tm:sys:provision:provisionstate'


class Gtm(UnnamedResource):
    """BIG-IP® system provision gtm resource

    The afm object only supports load and update because it is an
    unnamed resource.

    .. note::

        This is an unnamed resource so it has not ~Partition~Name pattern
        at the end of its URI.
    """
    def __init__(self, provision):
        super(Gtm, self).__init__(provision)
        self._meta_data['required_load_parameters'] = set()
        self._meta_data['required_json_kind'] =\
            'tm:sys:provision:provisionstate'


class Ili(UnnamedResource):
    """BIG-IP® system provision ili resource

    The afm object only supports load and update because it is an
    unnamed resource.

    .. note::

        This is an unnamed resource so it has not ~Partition~Name pattern
        at the end of its URI.
    """
    def __init__(self, provision):
        super(Ili, self).__init__(provision)
        self._meta_data['required_load_parameters'] = set()
        self._meta_data['required_json_kind'] =\
            'tm:sys:provision:provisionstate'


class Ilx(UnnamedResource):
    """BIG-IP® system provision ilx resource

    .. note::

        This is an unnamed resource so it has not ~Partition~Name pattern
        at the end of its URI.
    """
    def __init__(self, provision):
        super(Ilx, self).__init__(provision)
        self._meta_data['required_load_parameters'] = set()
        self._meta_data['required_json_kind'] =\
            'tm:sys:provision:provisionstate'


class Lc(UnnamedResource):
    """BIG-IP® system provision lc resource

    The afm object only supports load and update because it is an
    unnamed resource.

    .. note::

        This is an unnamed resource so it has not ~Partition~Name pattern
        at the end of its URI.
    """
    def __init__(self, provision):
        super(Lc, self).__init__(provision)
        self._meta_data['required_load_parameters'] = set()
        self._meta_data['required_json_kind'] =\
            'tm:sys:provision:provisionstate'


class Ltm(UnnamedResource):
    """BIG-IP® system provision ltm resource

    The afm object only supports load and update because it is an
    unnamed resource.

    .. note::

        This is an unnamed resource so it has not ~Partition~Name pattern
        at the end of its URI.
    """
    def __init__(self, provision):
        super(Ltm, self).__init__(provision)
        self._meta_data['required_load_parameters'] = set()
        self._meta_data['required_json_kind'] =\
            'tm:sys:provision:provisionstate'


class Pem(UnnamedResource):
    """BIG-IP® system provision pem resource

    The afm object only supports load and update because it is an
    unnamed resource.

    .. note::

        This is an unnamed resource so it has not ~Partition~Name pattern
        at the end of its URI.
    """
    def __init__(self, provision):
        super(Pem, self).__init__(provision)
        self._meta_data['required_load_parameters'] = set()
        self._meta_data['required_json_kind'] =\
            'tm:sys:provision:provisionstate'


class Swg(UnnamedResource):
    """BIG-IP® system provision swg resource

    The afm object only supports load and update because it is an
    unnamed resource.

    .. note::

        This is an unnamed resource so it has not ~Partition~Name pattern
        at the end of its URI.
    """
    def __init__(self, provision):
        super(Swg, self).__init__(provision)
        self._meta_data['required_load_parameters'] = set()
        self._meta_data['required_json_kind'] =\
            'tm:sys:provision:provisionstate'


class Urldb(UnnamedResource):
    """BIG-IP® system provision urldb resource

    The afm object only supports load and update because it is an
    unnamed resource.

    .. note::

        This is an unnamed resource so it has not ~Partition~Name pattern
        at the end of its URI.
    """
    def __init__(self, provision):
        super(Urldb, self).__init__(provision)
        self._meta_data['required_load_parameters'] = set()
        self._meta_data['required_json_kind'] =\
            'tm:sys:provision:provisionstate'


class Vcmp(UnnamedResource):
    """BIG-IP® system provision vcmp resource

    The vcmp object only supports load and update because it is an
    unnamed resource.

    .. note::

        This is an unnamed resource so it has not ~Partition~Name pattern
        at the end of its URI.
    """
    def __init__(self, provision):
        super(Vcmp, self).__init__(provision)
        self._meta_data['required_load_parameters'] = set()
        self._meta_data['required_json_kind'] =\
            'tm:sys:provision:provisionstate'



# coding=utf-8
#
# Copyright 2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""BIG-IP® utility module

REST URI
    ``http://localhost/mgmt/tm/util/unix-ls``

GUI Path
    N/A

REST Kind
    ``tm:util:unix-ls:*``
"""

from f5.bigip.mixins import CommandExecutionMixin
from f5.bigip.resource import UnnamedResource


class Unix_Ls(UnnamedResource, CommandExecutionMixin):

    def __init__(self, util):
        super(Unix_Ls, self).__init__(util)
        self._meta_data['required_command_parameters'].update(('utilCmdArgs',))
        self._meta_data['required_json_kind'] = 'tm:util:unix-ls:runstate'
        self._meta_data['allowed_commands'].append('run')



# coding=utf-8
#
# Copyright 2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""BIG-IP® Utility (util) module

REST URI
    ``http://localhost/mgmt/tm/util/``

GUI Path
    ``System``

REST Kind
    N/A -- HTTP GET returns an error
"""

from f5.bigip.resource import PathElement
from f5.bigip.tm.util.bash import Bash
from f5.bigip.tm.util.clientssl_ciphers import Clientssl_Ciphers
from f5.bigip.tm.util.dig import Dig
from f5.bigip.tm.util.get_dossier import Get_Dossier
from f5.bigip.tm.util.qkview import Qkview
from f5.bigip.tm.util.serverssl_ciphers import Serverssl_Ciphers
from f5.bigip.tm.util.unix_ls import Unix_Ls
from f5.bigip.tm.util.unix_mv import Unix_Mv
from f5.bigip.tm.util.unix_rm import Unix_Rm


class Util(PathElement):
    def __init__(self, bigip):
        super(Util, self).__init__(bigip)
        self._meta_data['allowed_lazy_attributes'] = [
            Bash,
            Clientssl_Ciphers,
            Dig,
            Get_Dossier,
            Qkview,
            Serverssl_Ciphers,
            Unix_Ls,
            Unix_Mv,
            Unix_Rm
        ]



# coding=utf-8
#
"""Classes and functions for configuring BIG-IP"""
# Copyright 2014 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


from icontrol.session import iControlRESTSession
try:
    import urlparse
except ImportError:
    import urllib.parse as urlparse

try:
    import signal
    from signal import SIGALRM
    HAS_SIGNAL = True
except ImportError:
    HAS_SIGNAL = False

from f5.bigip.cm import Cm
from f5.bigip.resource import PathElement
from f5.bigip.shared import Shared
from f5.bigip.tm.auth import Auth as TmAuth
from f5.bigip.tm.cm import Cm as TmCm
from f5.bigip.tm.gtm import Gtm
from f5.bigip.tm.ltm import Ltm
from f5.bigip.tm.net import Net
from f5.bigip.tm.shared import Shared as TmShared
from f5.bigip.tm.sys import Sys
from f5.bigip.tm import Tm
from f5.bigip.tm.transaction import Transactions
from f5.sdk_exception import TimeoutError


def timeout_handler(signum, frame):
    raise TimeoutError("Timed out waiting for response")


class BaseManagement(PathElement):
    def __init__(self, hostname, username, password, **kwargs):
        self.args = self.parse_arguments(
            hostname, username, password, **kwargs
        )
        self.icrs = self._get_icr_session(**self.args)
        self.configure_meta_data(**self.args)
        self.set_icr_metadata(self.icrs)

    def parse_arguments(self, *args, **kwargs):
        result = dict(
            timeout=kwargs.pop('timeout', 30),
            port=kwargs.pop('port', 443),
            icontrol_version=kwargs.pop('icontrol_version', ''),
            token=kwargs.pop('token', False),
            verify=kwargs.pop('verify', False),
            auth_provider=kwargs.pop('auth_provider', None),
            debug=kwargs.pop('debug', False)
        )
        if kwargs:
            raise TypeError('Unexpected **kwargs: %r' % kwargs)
        result.update(dict(
            hostname=args[0],
            username=args[1],
            password=args[2]
        ))
        return result

    def _get_icr_session(self, *args, **kwargs):
        params = dict(
            username=kwargs['username'],
            password=kwargs['password'],
            timeout=kwargs['timeout'],
            verify=kwargs['verify']
        )
        if kwargs['auth_provider']:
            params['auth_provider'] = kwargs['auth_provider']
        else:
            params['token'] = kwargs['token']

        result = iControlRESTSession(**params)
        result.debug = kwargs['debug']
        return result

    def configure_meta_data(self, *args, **kwargs):
        self._meta_data = {
            'allowed_lazy_attributes': [Tm, Cm, Shared],
            'hostname': kwargs['hostname'],
            'port': kwargs['port'],
            'device_name': None,
            'local_ip': None,
            'bigip': self,
            'icontrol_version': kwargs['icontrol_version'],
            'username': kwargs['username'],
            'password': kwargs['password'],
            'tmos_version': None,
        }

    def set_icr_metadata(self, icrs):
        self._meta_data['icr_session'] = icrs

    def post_configuration_setup(self):
        self._get_tmos_version()

    def _get_tmos_version(self):
        connect = self._meta_data['bigip']._meta_data['icr_session']
        base_uri = self._meta_data['uri'] + 'tm/sys/'
        if HAS_SIGNAL:
            signal.signal(SIGALRM, timeout_handler)
            signal.alarm(int(self.args['timeout']))
            response = connect.get(base_uri)
            signal.alarm(0)
        else:
            response = connect.get(base_uri)
        ver = response.json()
        version = urlparse.parse_qs(
            urlparse.urlparse(ver['selfLink']).query)['ver'][0]
        self._meta_data['tmos_version'] = version

    @property
    def hostname(self):
        return self._meta_data['hostname']

    @property
    def icontrol_version(self):
        return self._meta_data['icontrol_version']

    @property
    def tmos_version(self):
        if self._meta_data['tmos_version'] is None:
            self._meta_data['tmos_version'] = self._get_tmos_version()
        return self._meta_data['tmos_version']

    @property
    def debug(self):
        return self.icrs.debug

    @debug.setter
    def debug(self, value):
        self.icrs.debug = value

    @property
    def debug_output(self):
        result = []
        if self.icrs.debug_output:
            result += self.icrs.debug_output
        return result


class ManagementRoot(BaseManagement):
    """An interface to a single BIG-IP"""
    def __init__(self, hostname, username, password, **kwargs):
        super(ManagementRoot, self).__init__(
            hostname, username, password, **kwargs
        )
        self.set_metadata_uri(**self.args)
        self.post_configuration_setup()

    def set_metadata_uri(self, *args, **kwargs):
        self._meta_data['uri'] = 'https://{0}:{1}/mgmt/'.format(
            kwargs['hostname'], kwargs['port']
        )


class BigIP(ManagementRoot):
    """A shim class used to access the default config resources in 'mgmt/tm.'

    PLEASE DO NOT ADD ATTRIBUTES TO THIS CLASS.

    This class is depcrated in favor of MangementRoot above. Do not add any
    more objects to the allowed_lazy_attributes list here!

    This class is solely implemented for backwards compatibility.
    """
    def __init__(self, hostname, username, password, **kwargs):
        super(BigIP, self).__init__(hostname, username, password, **kwargs)
        self._meta_data['uri'] = self._meta_data['uri'] + 'tm/'
        self._meta_data['allowed_lazy_attributes'] =\
            [TmAuth, TmCm, Ltm, Gtm, Net, TmShared, Sys, Transactions]



# coding=utf-8
# Copyright 2016 F5 Networks Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#

'''The classes within define the management of a cluster of BIG_IP devices.

Definitions:
    Cluster: The manager of the TrustDomain and DeviceGroup objects.
    TrustDomain: a group of BIG-IP® devices that have exchanged certificates
                 and trust one another
    DeviceGroup: a group of BIG-IP® device that sync configuration data and
                 failover connections.

Clustering is broken down into three component parts: a cluster manager, a
trust domain, and a device group. The cluster manager presents the external
interface to a user for operations like create, teardown etc....

To create a device service group (aka cluster) of devices, those devices
must trust one another. This is coordinated by the TrustDomain class. Once
those devices trust one another, a device group is created and each is added
to the group. After this step, a cluster exists.

Currently the only supported type of cluster is a 'sync-failover' cluster.
The number of devices supported officially is currently two, for an
active-standby cluster, but the code below can accommodate a four-member
cluster.

Methods:

    * create -- creates a cluster based on kwargs given by user
    * teardown -- tears down an existing cluster

Examples:

There are two major use-cases here:

    * Manage an existing cluster:

        list_of_bigips = [ManagementRoot(...), ManagementRoot(...)]
        cluster_mgr = ClusterManager(
                            devices=list_of_bigips,
                            device_group_name='my_cluster',
                            device_group_type='sync-failover',
                            device_group_partition='Common'
                        )
        assert cluster_mgr.cluster.devices == list_of_bigips

    * Create a new cluster and manage it:

        list_of_bigips = [ManagementRoot(...), ManagementRoot(...)]
        cluster_mgr = ClusterManager()
        cluster_mgr.create(
                    devices=list_of_bigips,
                    device_group_name='my_cluster',
                    device_group_type='sync-failover',
                    device_group_partition='Common
                )
        assert cluster_mgr.cluster.devices == list_of_bigips
'''

from f5.multi_device.device_group import DeviceGroup
from f5.multi_device.trust_domain import TrustDomain

from f5.multi_device.exceptions import AlreadyManagingCluster
from f5.multi_device.exceptions import ClusterNotSupported
from f5.multi_device.exceptions import NoClusterToManage


from collections import namedtuple


Cluster = namedtuple(
    'Cluster', 'device_group_name device_group_type device_group_partition '
    'devices'
)


class ClusterManager(object):
    '''Manage a cluster of BIG-IP® devices.

    This is accomplished with REST URI calls only, but some operations are
    only permitted via tmsh commands (such as adding cm/trust-domain peers).
    We get around this issue by deploying iApps (sys/application).
    '''

    def __init__(self, **kwargs):
        '''Initialize a cluster manager object.

        :param bigips: list -- list of bigip.ManagementRoot ojects to cluster
        :param cluster_name: str -- name of device service group
        :param partition: str -- partition to associate cluster with
        :param cluster_type: str -- type of cluster configuration
        '''

        if kwargs:
            self.manage_extant(**kwargs)
        else:
            self.trust_domain = TrustDomain()
            self.device_group = DeviceGroup()

    def __getattr__(self, name):
        '''If ClusterManager.cluster is not found come here.

        :param name: str -- name of attribute
        :raises: AttributeError
        '''

        if name == 'cluster':
            msg = 'The ClusterManager is not managing a cluster.'
            raise NoClusterToManage(msg)
        raise AttributeError(name)

    def _check_device_number(self, devices):
        '''Check if number of devices is between 2 and 4

        :param kwargs: dict -- keyword args in dict
        '''

        if len(devices) < 2 or len(devices) > 4:
            msg = 'The number of devices to cluster is not supported.'
            raise ClusterNotSupported(msg)

    def manage_extant(self, **kwargs):
        '''Manage an existing cluster

        :param kwargs: dict -- keyword args in dict
        '''

        self._check_device_number(kwargs['devices'])
        self.trust_domain = TrustDomain(
            devices=kwargs['devices'],
            partition=kwargs['device_group_partition']
        )
        self.device_group = DeviceGroup(**kwargs)
        self.cluster = Cluster(**kwargs)

    def create(self, **kwargs):
        '''Create a cluster of BIG-IP® devices.

        :param kwargs: dict -- keyword arguments for cluster manager
        '''

        try:
            cluster = getattr(self, "cluster", None)
        except NoClusterToManage:
            cluster = None
        if cluster is not None:
            msg = 'The ClusterManager is already managing a cluster.'
            raise AlreadyManagingCluster(msg)

        self._check_device_number(kwargs['devices'])
        self.trust_domain.create(
            devices=kwargs['devices'],
            partition=kwargs['device_group_partition']
        )
        self.device_group.create(**kwargs)
        self.cluster = Cluster(**kwargs)

    def teardown(self):
        '''Teardown the cluster of BIG-IP® devices.'''

        self.device_group.teardown()
        self.trust_domain.teardown()
        self.cluster = None



# coding=utf-8
#
# Copyright 2017 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""BIG-IQ® license pool regkeys.

REST URI
    ``http://localhost/mgmt/cm/device/licensing/pool/regkey/licenses``

REST Kind
    ``cm:device:licensing:pool:regkey:licenses:*``
"""

import os
import time
import uuid

from f5.bigiq.resource import Collection
from f5.bigiq.resource import OrganizingCollection
from f5.bigiq.resource import Resource
from f5.sdk_exception import F5SDKError
from f5.sdk_exception import RequiredOneOf
from six import iterkeys


class Regkey(OrganizingCollection):
    def __init__(self, pool):
        super(Regkey, self).__init__(pool)
        self._meta_data['allowed_lazy_attributes'] = [
            Licenses_s
        ]


class Licenses_s(Collection):
    def __init__(self, regkey):
        super(Licenses_s, self).__init__(regkey)
        self._meta_data['required_json_kind'] = \
            'cm:device:licensing:pool:regkey:licenses:regkeypoollicensecollectionstate'  # NOQA
        self._meta_data['allowed_lazy_attributes'] = [Licenses]
        self._meta_data['attribute_registry'] = {
            'cm:device:licensing:pool:regkey:licenses:regkeypoollicensestate': Licenses  # NOQA
        }


class Licenses(Resource):
    def __init__(self, licenses_s):
        super(Licenses, self).__init__(licenses_s)
        self._meta_data['required_creation_parameters'] = {'name', }
        self._meta_data['required_json_kind'] = \
            'cm:device:licensing:pool:regkey:licenses:regkeypoollicensestate'
        self._meta_data['allowed_lazy_attributes'] = [
            Offerings_s
        ]
        self._meta_data['attribute_registry'] = {
            'cm:device:licensing:pool:regkey:licenses:item:offerings:regkeypoollicenseofferingcollectionstate': Offerings_s  # NOQA
        }


class Offerings_s(Collection):
    def __init__(self, license):
        super(Offerings_s, self).__init__(license)
        self._meta_data['required_json_kind'] = \
            'cm:device:licensing:pool:regkey:licenses:item:offerings:regkeypoollicenseofferingcollectionstate'  # NOQA
        self._meta_data['allowed_lazy_attributes'] = [Offerings]
        self._meta_data['attribute_registry'] = {
            'cm:device:licensing:pool:regkey:licenses:item:offerings:regkeypoollicenseofferingstate': Offerings  # NOQA
        }


class Offerings(Resource):
    def __init__(self, offerings_s):
        super(Offerings, self).__init__(offerings_s)
        self._meta_data['required_creation_parameters'] = {'regKey', }
        self._meta_data['required_json_kind'] = \
            'cm:device:licensing:pool:regkey:licenses:item:offerings:regkeypoollicenseofferingstate'  # NOQA
        self._meta_data['allowed_lazy_attributes'] = [Members_s]
        self._meta_data['attribute_registry'] = {
            'cm:device:licensing:pool:regkey:licenses:item:offerings:regkey:members:regkeypoollicensemembercollectionstate': Members_s
        }


class Members_s(Collection):
    def __init__(self, pool):
        super(Members_s, self).__init__(pool)
        self._meta_data['required_json_kind'] = \
            'cm:device:licensing:pool:regkey:licenses:item:offerings:regkey:members:regkeypoollicensemembercollectionstate'  # NOQA
        self._meta_data['allowed_lazy_attributes'] = [Members]
        self._meta_data['attribute_registry'] = {
            'cm:device:licensing:pool:regkey:licenses:item:offerings:regkey:members:regkeypoollicensememberstate': Members
        }


class Members(Resource):
    def __init__(self, members_s):
        super(Members, self).__init__(members_s)
        self._meta_data['required_json_kind'] = 'cm:device:licensing:pool:regkey:licenses:item:offerings:regkey:members:regkeypoollicensememberstate'

        # This set is empty because the creation checking is done as
        # a required_one_of in the create() method
        self._meta_data['required_creation_parameters'] = set()

    def create(self, **kwargs):
        required_one_of = [
            ['deviceAddress', 'username', 'password'],
            ['deviceReference']
        ]

        args = set(iterkeys(kwargs))

        # Check to see if the user supplied any of the sets of required
        # arguments.
        has_any = [set(x).issubset(args) for x in required_one_of]

        if len([x for x in has_any if x is True]) == 1:
            # Only one of the required argument sets can be provided
            # so if more than that are provided, it is an error.
            return self._create(**kwargs)

        raise RequiredOneOf(required_one_of)

    def delete(self, **kwargs):
        """Deletes a member from a license pool

        You need to be careful with this method. When you use it, and it
        succeeds on the remote BIG-IP, the configuration of the BIG-IP
        will be reloaded. During this process, you will not be able to
        access the REST interface.

        This method overrides the Resource class's method because it requires
        that extra json kwargs be supplied. This is not a behavior that is
        part of the normal Resource class's delete method.

        :param kwargs:
        :return:
        """
        if 'id' not in kwargs:
            # BIG-IQ requires that you provide the ID of the members to revoke
            # a license from. This ID is already part of the deletion URL though.
            # Therefore, if you do not provide it, we enumerate it for you.
            delete_uri = self._meta_data['uri']
            if delete_uri.endswith('/'):
                delete_uri = delete_uri[0:-1]
                kwargs['id'] = os.path.basename(delete_uri)
            uid = uuid.UUID(kwargs['id'], version=4)
            if uid.hex != kwargs['id'].replace('-', ''):
                raise F5SDKError(
                    "The specified ID is invalid"
                )

        requests_params = self._handle_requests_params(kwargs)
        kwargs = self._check_for_python_keywords(kwargs)
        kwargs = self._prepare_request_json(kwargs)

        delete_uri = self._meta_data['uri']
        session = self._meta_data['bigip']._meta_data['icr_session']

        # Check the generation for match before delete
        force = self._check_force_arg(kwargs.pop('force', True))
        if not force:
            self._check_generation()

        response = session.delete(delete_uri, json=kwargs, **requests_params)
        if response.status_code == 200:
            self.__dict__ = {'deleted': True}

        # This sleep is necessary to prevent BIG-IQ from being able to remove
        # a license. It happens in certain cases that assignments can be revoked
        # (and license deletion started) too quickly. Therefore, we must introduce
        # an artificial delay here to prevent revoking from returning before
        # BIG-IQ would be ready to remove the license.
        time.sleep(1)



# coding=utf-8
#
#  Copyright 2014-2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""BIG-IP® LTM monitor submodule.

REST URI
    ``http://localhost/mgmt/tm/ltm/monitor/``

GUI Path
    ``Local Traffic --> Monitors``

REST Kind
    ``tm:ltm:monitor*``
"""

from f5.bigip.mixins import UpdateMonitorMixin
from f5.bigip.resource import Collection
from f5.bigip.resource import OrganizingCollection
from f5.bigip.resource import Resource


class Monitor(OrganizingCollection):
    def __init__(self, ltm):
        super(Monitor, self).__init__(ltm)
        self._meta_data['allowed_lazy_attributes'] = [
            Https,
            Https_s,
            Diameters,
            Dns_s,
            Externals,
            Firepass_s,
            Ftps,
            Gateway_Icmps,
            Icmps,
            Imaps,
            Inbands,
            Ldaps,
            Module_Scores,
            Mqtts,
            Mssqls,
            Mysqls,
            Nntps,
            Oracles,
            Pop3s,
            Postgresqls,
            Radius_s,
            Radius_Accountings,
            Real_Servers,
            Rpcs,
            Sasps,
            Scripteds,
            Sips,
            Smbs,
            Smtps,
            Snmp_Dcas,
            Snmp_Dca_Bases,
            Soaps,
            Tcps,
            Tcp_Echos,
            Tcp_Half_Opens,
            Udps,
            Virtual_Locations,
            Waps,
            Wmis]


class Https(Collection):
    """BIG-IP® Http monitor collection."""
    def __init__(self, monitor):
        super(Https, self).__init__(monitor)
        self._meta_data['allowed_lazy_attributes'] = [Http]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:http:httpstate': Http}


class Http(UpdateMonitorMixin, Resource):
    """BIG-IP® Http monitor resource."""
    def __init__(self, https):
        super(Http, self).__init__(https)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:http:httpstate'


class Https_s(Collection):
    """BIG-IP® Https monitor collection."""
    def __init__(self, monitor):
        super(Https_s, self).__init__(monitor)
        self._meta_data['allowed_lazy_attributes'] = [HttpS]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:https:httpsstate': HttpS}


class HttpS(UpdateMonitorMixin, Resource):
    """BIG-IP® Https monitor resource."""
    def __init__(self, https_s):
        super(HttpS, self).__init__(https_s)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:https:httpsstate'


class Diameters(Collection):
    """BIG-IP® diameter monitor collection."""
    def __init__(self, monitor):
        super(Diameters, self).__init__(monitor)
        self._meta_data['allowed_lazy_attributes'] = [Diameter]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:diameter:diameterstate': Diameter}


class Diameter(UpdateMonitorMixin, Resource):
    """BIG-IP® diameter monitor resource."""
    def __init__(self, diameters):
        super(Diameter, self).__init__(diameters)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:diameter:diameterstate'


class Dns_s(Collection):
    """BIG-IP® Dns monitor collection."""
    def __init__(self, monitor):
        super(Dns_s, self).__init__(monitor)
        self._meta_data['allowed_lazy_attributes'] = [Dns]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:dns:dnsstate': Dns}


class Dns(UpdateMonitorMixin, Resource):
    """BIG-IP® Dns monitor resource."""
    def __init__(self, dns_s):
        super(Dns, self).__init__(dns_s)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:dns:dnsstate'
        self._meta_data['required_creation_parameters'].update(('qname',))


class Externals(Collection):
    """BIG-IP® external monitor collection."""
    def __init__(self, monitor):
        super(Externals, self).__init__(monitor)
        self._meta_data['allowed_lazy_attributes'] = [External]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:external:externalstate': External}


class External(UpdateMonitorMixin, Resource):
    """BIG-IP® external monitor resrouce."""
    def __init__(self, externals):
        super(External, self).__init__(externals)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:external:externalstate'


class Firepass_s(Collection):
    """BIG-IP® Fire Pass monitor collection."""
    def __init__(self, monitor):
        super(Firepass_s, self).__init__(monitor)
        self._meta_data['allowed_lazy_attributes'] = [Firepass]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:firepass:firepassstate': Firepass}


class Firepass(UpdateMonitorMixin, Resource):
    """BIG-IP® external monitor resource."""
    def __init__(self, firepass_s):
        super(Firepass, self).__init__(firepass_s)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:firepass:firepassstate'


class Ftps(Collection):
    """BIG-IP® Ftp monitor collection."""
    def __init__(self, monitor):
        super(Ftps, self).__init__(monitor)
        self._meta_data['allowed_lazy_attributes'] = [Ftp]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:ftp:ftpstate': Ftp}


class Ftp(UpdateMonitorMixin, Resource):
    """BIG-IP® Ftp monitor resource."""
    def __init__(self, ftps):
        super(Ftp, self).__init__(ftps)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:ftp:ftpstate'


class Gateway_Icmps(Collection):
    """BIG-IP® Gateway Icmp monitor collection."""
    def __init__(self, monitor):
        super(Gateway_Icmps, self).__init__(monitor)
        fixed = self._meta_data['uri'].replace('_', '-')
        self._meta_data['uri'] = fixed
        self._meta_data['allowed_lazy_attributes'] = [Gateway_Icmp]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:gateway-icmp:gateway-icmpstate': Gateway_Icmp}


class Gateway_Icmp(UpdateMonitorMixin, Resource):
    """BIG-IP® Gateway Icmp monitor resource."""
    def __init__(self, gateway_icmps):
        super(Gateway_Icmp, self).__init__(gateway_icmps)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:gateway-icmp:gateway-icmpstate'


class Icmps(Collection):
    """BIG-IP® Icmp monitor collection."""
    def __init__(self, monitor):
        super(Icmps, self).__init__(monitor)
        self._meta_data['allowed_lazy_attributes'] = [Icmp]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:icmp:icmpstate': Icmp}


class Icmp(UpdateMonitorMixin, Resource):
    """BIG-IP® Icmp monitor resource."""
    def __init__(self, icmps):
        super(Icmp, self).__init__(icmps)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:icmp:icmpstate'


class Imaps(Collection):
    """BIG-IP® Imap monitor collection."""
    def __init__(self, monitor):
        super(Imaps, self).__init__(monitor)
        self._meta_data['allowed_lazy_attributes'] = [Imap]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:imap:imapstate': Imap}


class Imap(UpdateMonitorMixin, Resource):
    """BIG-IP® Imap monitor resource."""
    def __init__(self, imaps):
        super(Imap, self).__init__(imaps)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:imap:imapstate'


class Inbands(Collection):
    """BIG-IP® in band monitor collection."""
    def __init__(self, monitor):
        super(Inbands, self).__init__(monitor)
        self._meta_data['allowed_lazy_attributes'] = [Inband]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:inband:inbandstate': Inband}


class Inband(UpdateMonitorMixin, Resource):
    """BIG-IP® in band monitor resource."""
    def __init__(self, inbands):
        super(Inband, self).__init__(inbands)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:inband:inbandstate'


class Ldaps(Collection):
    """BIG-IP® Ldap monitor collection."""
    def __init__(self, monitor):
        super(Ldaps, self).__init__(monitor)
        self._meta_data['allowed_lazy_attributes'] = [Ldap]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:ldap:ldapstate': Ldap}


class Ldap(UpdateMonitorMixin, Resource):
    """BIG-IP® Ldap monitor resource."""
    def __init__(self, ldaps):
        super(Ldap, self).__init__(ldaps)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:ldap:ldapstate'


class Module_Scores(Collection):
    """BIG-IP® module scores monitor collection."""
    def __init__(self, monitor):
        super(Module_Scores, self).__init__(monitor)
        fixed = self._meta_data['uri'].replace('_', '-')
        self._meta_data['uri'] = fixed
        self._meta_data['allowed_lazy_attributes'] = [Module_Score]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:module-score:module-scorestate': Module_Score}


class Module_Score(UpdateMonitorMixin, Resource):
    """BIG-IP® module scores monitor resource."""
    def __init__(self, gateway_icmps):
        super(Module_Score, self).__init__(gateway_icmps)
        self._meta_data['required_creation_parameters'].update(
            ('snmpIpAddress',))
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:module-score:module-scorestate'


class Mysqls(Collection):
    """BIG-IP® MySQL monitor collection."""
    def __init__(self, monitor):
        super(Mysqls, self).__init__(monitor)
        self._meta_data['allowed_lazy_attributes'] = [Mysql]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:mysql:mysqlstate': Mysql}


class Mysql(UpdateMonitorMixin, Resource):
    """BIG-IP® MySQL monitor resource."""
    def __init__(self, mysqls):
        super(Mysql, self).__init__(mysqls)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:mysql:mysqlstate'


class Mqtts(Collection):
    def __init__(self, monitor):
        super(Mqtts, self).__init__(monitor)
        self._meta_data['allowed_lazy_attributes'] = [Mqtt]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:mqtt:mqttstate': Mqtt}


class Mqtt(UpdateMonitorMixin, Resource):
    def __init__(self, mqtts):
        super(Mqtt, self).__init__(mqtts)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:mqtt:mqttstate'


class Mssqls(Collection):
    """BIG-IP® Mssql monitor collection."""
    def __init__(self, monitor):
        super(Mssqls, self).__init__(monitor)
        self._meta_data['allowed_lazy_attributes'] = [Mssql]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:mssql:mssqlstate': Mssql}


class Mssql(UpdateMonitorMixin, Resource):
    """BIG-IP® Mssql monitor resource."""
    def __init__(self, mssqls):
        super(Mssql, self).__init__(mssqls)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:mssql:mssqlstate'


class Nntps(Collection):
    """BIG-IP® Nntps monitor collection."""
    def __init__(self, monitor):
        super(Nntps, self).__init__(monitor)
        self._meta_data['allowed_lazy_attributes'] = [Nntp]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:nntp:nntpstate': Nntp}


class Nntp(UpdateMonitorMixin, Resource):
    """BIG-IP® Nntps monitor resource."""
    def __init__(self, nntps):
        super(Nntp, self).__init__(nntps)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:nntp:nntpstate'


class Nones(Collection):
    """BIG-IP® None monitor collection."""
    def __init__(self, monitor):
        super(Nones, self).__init__(monitor)
        self._meta_data['allowed_lazy_attributes'] = [NONE]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:none:nonestate': NONE}


class NONE(UpdateMonitorMixin, Resource):
    """BIG-IP® None monitor resource."""
    def __init__(self, nones):
        super(NONE, self).__init__(nones)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:none:nonestate'


class Oracles(Collection):
    """BIG-IP® Oracle monitor collection."""
    def __init__(self, monitor):
        super(Oracles, self).__init__(monitor)
        self._meta_data['allowed_lazy_attributes'] = [Oracle]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:oracle:oraclestate': Oracle}


class Oracle(UpdateMonitorMixin, Resource):
    """BIG-IP® Oracle monitor resource."""
    def __init__(self, oracles):
        super(Oracle, self).__init__(oracles)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:oracle:oraclestate'


class Pop3s(Collection):
    """BIG-IP® Pop3 monitor collection."""
    def __init__(self, monitor):
        super(Pop3s, self).__init__(monitor)
        self._meta_data['allowed_lazy_attributes'] = [Pop3]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:pop3:pop3state': Pop3}


class Pop3(UpdateMonitorMixin, Resource):
    """BIG-IP® Pop3 monitor resource."""
    def __init__(self, pop3s):
        super(Pop3, self).__init__(pop3s)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:pop3:pop3state'


class Postgresqls(Collection):
    """BIG-IP® PostGRES SQL monitor collection."""
    def __init__(self, monitor):
        super(Postgresqls, self).__init__(monitor)
        self._meta_data['allowed_lazy_attributes'] = [Postgresql]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:postgresql:postgresqlstate': Postgresql}


class Postgresql(UpdateMonitorMixin, Resource):
    """BIG-IP® PostGRES SQL monitor resource."""
    def __init__(self, postgresqls):
        super(Postgresql, self).__init__(postgresqls)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:postgresql:postgresqlstate'


class Radius_s(Collection):
    """BIG-IP® radius monitor collection."""
    def __init__(self, monitor):
        super(Radius_s, self).__init__(monitor)
        self._meta_data['allowed_lazy_attributes'] = [Radius]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:radius:radiusstate': Radius}


class Radius(UpdateMonitorMixin, Resource):
    """BIG-IP® radius monitor resource."""
    def __init__(self, radius_s):
        super(Radius, self).__init__(radius_s)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:radius:radiusstate'


class Radius_Accountings(Collection):
    """BIG-IP® radius accounting monitor collection."""
    def __init__(self, monitor):
        super(Radius_Accountings, self).__init__(monitor)
        fixed = self._meta_data['uri'].replace('_', '-')
        self._meta_data['uri'] = fixed
        self._meta_data['allowed_lazy_attributes'] = [Radius_Accounting]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:radius-accounting:radius-accountingstate':
             Radius_Accounting}


class Radius_Accounting(UpdateMonitorMixin, Resource):
    """BIG-IP® radius accounting monitor resource."""
    def __init__(self, radius_accountings):
        super(Radius_Accounting, self).__init__(radius_accountings)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:radius-accounting:radius-accountingstate'


class Real_Servers(Collection):
    """BIG-IP® real-server monitor collection."""
    def __init__(self, monitor):
        super(Real_Servers, self).__init__(monitor)
        fixed = self._meta_data['uri'].replace('_', '-')
        self._meta_data['uri'] = fixed
        self._meta_data['allowed_lazy_attributes'] = [Real_Server]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:real-server:real-serverstate': Real_Server}


class Real_Server(UpdateMonitorMixin, Resource):
    """BIG-IP® real-server monitor resource."""
    def __init__(self, real_servers):
        super(Real_Server, self).__init__(real_servers)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:real-server:real-serverstate'

    def update(self, **kwargs):
        """Change the configuration of the resource on the device.

        This method uses Http PUT alter the service state on the device.

        The attributes of the instance will be packaged as a dictionary.  That
        dictionary will be updated with kwargs.  It is then submitted as JSON
        to the device.  Various edge cases are handled:

        * read-only attributes that are unchangeable are removed
        * ``tmCommand`` attribute removed prior to PUT
        * ``agent`` attribute removed prior to PUT
        * ``post`` attribute removed prior to PUT


        :param kwargs: keys and associated values to alter on the device

        """
        self.__dict__.pop('tmCommand', '')
        self.__dict__.pop('agent', '')
        self.__dict__.pop('method', '')
        super(Real_Server, self).update(**kwargs)


class Rpcs(Collection):
    """BIG-IP® Rpc monitor collection."""
    def __init__(self, monitor):
        super(Rpcs, self).__init__(monitor)
        self._meta_data['allowed_lazy_attributes'] = [Rpc]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:rpc:rpcstate': Rpc}


class Rpc(UpdateMonitorMixin, Resource):
    """BIG-IP® Rpc monitor resource."""
    def __init__(self, rpcs):
        super(Rpc, self).__init__(rpcs)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:rpc:rpcstate'


class Sasps(Collection):
    """BIG-IP® Sasp monitor collection."""
    def __init__(self, monitor):
        super(Sasps, self).__init__(monitor)
        self._meta_data['allowed_lazy_attributes'] = [Sasp]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:sasp:saspstate': Sasp}


class Sasp(UpdateMonitorMixin, Resource):
    """BIG-IP® Sasp monitor resource."""
    def __init__(self, sasps):
        super(Sasp, self).__init__(sasps)
        self._meta_data['required_creation_parameters'].update(
            ('primaryAddress',))
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:sasp:saspstate'


class Scripteds(Collection):
    """BIG-IP® scripted monitor collection."""
    def __init__(self, monitor):
        super(Scripteds, self).__init__(monitor)
        self._meta_data['allowed_lazy_attributes'] = [Scripted]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:scripted:scriptedstate': Scripted}


class Scripted(UpdateMonitorMixin, Resource):
    """BIG-IP® scripted monitor resource."""
    def __init__(self, scripteds):
        super(Scripted, self).__init__(scripteds)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:scripted:scriptedstate'


class Sips(Collection):
    """BIG-IP® Sip monitor collection."""
    def __init__(self, monitor):
        super(Sips, self).__init__(monitor)
        self._meta_data['allowed_lazy_attributes'] = [Sip]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:sip:sipstate': Sip}


class Sip(UpdateMonitorMixin, Resource):
    """BIG-IP® Sip monitor resource."""
    def __init__(self, sips):
        super(Sip, self).__init__(sips)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:sip:sipstate'


class Smbs(Collection):
    """BIG-IP® Smb monitor collection."""
    def __init__(self, monitor):
        super(Smbs, self).__init__(monitor)
        self._meta_data['allowed_lazy_attributes'] = [Smb]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:smb:smbstate': Smb}


class Smb(UpdateMonitorMixin, Resource):
    """BIG-IP® Smb monitor resource."""
    def __init__(self, smbs):
        super(Smb, self).__init__(smbs)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:smb:smbstate'


class Smtps(Collection):
    """BIG-IP® Smtp monitor collection."""
    def __init__(self, monitor):
        super(Smtps, self).__init__(monitor)
        self._meta_data['allowed_lazy_attributes'] = [Smtp]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:smtp:smtpstate': Smtp}


class Smtp(UpdateMonitorMixin, Resource):
    """BIG-IP® Smtp monitor resource."""
    def __init__(self, smtps):
        super(Smtp, self).__init__(smtps)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:smtp:smtpstate'


class Snmp_Dcas(Collection):
    """BIG-IP® SNMP DCA monitor collection."""
    def __init__(self, monitor):
        super(Snmp_Dcas, self).__init__(monitor)
        fixed = self._meta_data['uri'].replace('_', '-')
        self._meta_data['uri'] = fixed
        self._meta_data['allowed_lazy_attributes'] = [Snmp_Dca]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:snmp-dca:snmp-dcastate': Snmp_Dca}


class Snmp_Dca(UpdateMonitorMixin, Resource):
    """BIG-IP® SNMP DCA monitor resource."""
    def __init__(self, snmp_dcas):
        super(Snmp_Dca, self).__init__(snmp_dcas)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:snmp-dca:snmp-dcastate'


class Snmp_Dca_Bases(Collection):
    """BIG-IP® SNMP DCA bases monitor collection."""
    def __init__(self, monitor):
        super(Snmp_Dca_Bases, self).__init__(monitor)
        fixed = self._meta_data['uri'].replace('_', '-')
        self._meta_data['uri'] = fixed
        self._meta_data['allowed_lazy_attributes'] = [Snmp_Dca_Base]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:snmp-dca-base:snmp-dca-basestate': Snmp_Dca_Base}


class Snmp_Dca_Base(UpdateMonitorMixin, Resource):
    """BIG-IP® SNMP DCA monitor resource."""
    def __init__(self, snmp_dca_bases):
        super(Snmp_Dca_Base, self).__init__(snmp_dca_bases)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:snmp-dca-base:snmp-dca-basestate'


class Soaps(Collection):
    """BIG-IP® Soap monitor collection."""
    def __init__(self, monitor):
        super(Soaps, self).__init__(monitor)
        self._meta_data['allowed_lazy_attributes'] = [Soap]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:soap:soapstate': Soap}


class Soap(UpdateMonitorMixin, Resource):
    """BIG-IP® Soap monitor resource."""
    def __init__(self, soaps):
        super(Soap, self).__init__(soaps)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:soap:soapstate'


class Tcps(Collection):
    """BIG-IP® Tcp monitor collection."""
    def __init__(self, monitor):
        super(Tcps, self).__init__(monitor)
        self._meta_data['allowed_lazy_attributes'] = [Tcp]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:tcp:tcpstate': Tcp}


class Tcp(UpdateMonitorMixin, Resource):
    """BIG-IP® Tcp monitor resource."""
    def __init__(self, tcps):
        super(Tcp, self).__init__(tcps)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:tcp:tcpstate'


class Tcp_Echos(Collection):
    """BIG-IP® Tcp echo monitor collection."""
    def __init__(self, monitor):
        super(Tcp_Echos, self).__init__(monitor)
        fixed = self._meta_data['uri'].replace('_', '-')
        self._meta_data['uri'] = fixed
        self._meta_data['allowed_lazy_attributes'] = [Tcp_Echo]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:tcp-echo:tcp-echostate': Tcp_Echo}


class Tcp_Echo(UpdateMonitorMixin, Resource):
    """BIG-IP® Tcp echo monitor resource."""
    def __init__(self, tcp_echos):
        super(Tcp_Echo, self).__init__(tcp_echos)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:tcp-echo:tcp-echostate'


class Tcp_Half_Opens(Collection):
    """BIG-IP® Tcp half open monitor collection."""
    def __init__(self, monitor):
        super(Tcp_Half_Opens, self).__init__(monitor)
        fixed = self._meta_data['uri'].replace('_', '-')
        self._meta_data['uri'] = fixed
        self._meta_data['allowed_lazy_attributes'] = [Tcp_Half_Open]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:tcp-half-open:tcp-half-openstate': Tcp_Half_Open}


class Tcp_Half_Open(UpdateMonitorMixin, Resource):
    """BIG-IP® Tcp half open monitor resource."""
    def __init__(self, tcp_half_opens):
        super(Tcp_Half_Open, self).__init__(tcp_half_opens)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:tcp-half-open:tcp-half-openstate'


class Udps(Collection):
    """BIG-IP® Udp monitor collection."""
    def __init__(self, monitor):
        super(Udps, self).__init__(monitor)
        self._meta_data['allowed_lazy_attributes'] = [Udp]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:udp:udpstate': Udp}


class Udp(UpdateMonitorMixin, Resource):
    """BIG-IP® Udp monitor resource."""
    def __init__(self, udps):
        super(Udp, self).__init__(udps)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:udp:udpstate'


class Virtual_Locations(Collection):
    """BIG-IP® virtual-locations monitor collection."""
    def __init__(self, monitor):
        super(Virtual_Locations, self).__init__(monitor)
        fixed = self._meta_data['uri'].replace('_', '-')
        self._meta_data['uri'] = fixed
        self._meta_data['allowed_lazy_attributes'] = [Virtual_Location]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:virtual-location:virtual-locationstate':
             Virtual_Location}


class Virtual_Location(UpdateMonitorMixin, Resource):
    """BIG-IP® virtual-locations monitor resource."""
    def __init__(self, virtual_locations):
        super(Virtual_Location, self).__init__(virtual_locations)
        self._meta_data['required_creation_parameters'].update(
            ('pool',))
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:virtual-location:virtual-locationstate'


class Waps(Collection):
    """BIG-IP® Wap monitor collection."""
    def __init__(self, monitor):
        super(Waps, self).__init__(monitor)
        self._meta_data['allowed_lazy_attributes'] = [Wap]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:wap:wapstate': Wap}


class Wap(UpdateMonitorMixin, Resource):
    """BIG-IP® Wap monitor resource."""
    def __init__(self, waps):
        super(Wap, self).__init__(waps)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:wap:wapstate'


class Wmis(Collection):
    """BIG-IP® Wmi monitor collection."""
    def __init__(self, monitor):
        super(Wmis, self).__init__(monitor)
        self._meta_data['allowed_lazy_attributes'] = [Wmi]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:monitor:wmi:wmistate': Wmi}


class Wmi(UpdateMonitorMixin, Resource):
    """BIG-IP® Wmi monitor resource."""
    def __init__(self, wmis):
        super(Wmi, self).__init__(wmis)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:monitor:wmi:wmistate'
        self._meta_data['read_only_attributes'] =\
            ['agent', 'post', 'method']

    def update(self, **kwargs):
        """Change the configuration of the resource on the device.

        This method uses Http PUT alter the service state on the device.

        The attributes of the instance will be packaged as a dictionary.  That
        dictionary will be updated with kwargs.  It is then submitted as JSON
        to the device.  Various edge cases are handled:

        * read-only attributes that are unchangeable are removed
        * ``agent`` attribute removed prior to PUT
        * ``post`` attribute removed prior to PUT
        * ``method`` attribute removed prior to PUT

        :param kwargs: keys and associated values to alter on the device

        """
        self.__dict__.pop('agent', '')
        self.__dict__.pop('post', '')
        self.__dict__.pop('method', '')
        super(Wmi, self).update(**kwargs)
